import matplotlib.pyplot as plt
import pickle


def energy_plot(y_max, y_min):
    path = './data/pkl_data/rslt_data.pkl'
    if not path:
        print("rslt_data does not exist")
    else:
        with open(path, 'rb') as f:
            rslt_data = pickle.load(f)

        ndata = len(rslt_data)
        Emin = rslt_data['E_eV_atom'].min()

        # ---------- figure size
        plt.rcParams['figure.figsize'] = [8, 6]

        # ---------- axes
        plt.rcParams['axes.grid'] = True
        plt.rcParams['axes.linewidth'] = 1.5

        # ---------- ticks
        plt.rcParams['xtick.direction'] = 'in'
        plt.rcParams['ytick.direction'] = 'in'
        plt.rcParams['xtick.major.width'] = 1.0
        plt.rcParams['ytick.major.width'] = 1.0
        plt.rcParams['xtick.major.size'] = 8.0
        plt.rcParams['ytick.major.size'] = 8.0

        # ---------- lines
        plt.rcParams['lines.linewidth'] = 2.5

        # ---------- grid
        plt.rcParams['grid.linestyle'] = ':'

        # ---------- font
        plt.rcParams['font.size'] = 20
        # plt.rcParams['pdf.fonttype'] = 42    # embed fonts in PDF using type42 (True type)

        fig, ax = plt.subplots()

        # ---------- axis
        dx = 1
        ax.set_xlim([0, ndata+dx])
        ax.set_ylim([y_min, y_max])

        # ---------- hline at zero
        ax.hlines(0.0, -dx, ndata+dx, 'k', '--')

        # ---------- plot
        # x <-- ID + 1
        # ax.plot(rslt_data.index + 1,
        #        rslt_data['E_eV_atom'], 'o', ms=15, mew=2.0, alpha=0.8)

        ax.plot(rslt_data.index + 1,
                rslt_data['E_eV_atom']-Emin, 'o', ms=15, mew=2.0, alpha=0.8)

        # ---------- title and label
        # ax.set_title('Random search for Si$_{16}$')
        ax.set_xlabel('Number of trials')
        ax.set_ylabel('Energy (eV/atom)')



#!/usr/bin/env python3
'''
Main script
'''

import argparse
from logging import getLogger
import os

from cryspy.start import cryspy_init, cryspy_restart
from cryspy.util.utility import set_logger, backup_cryspy, clean_cryspy

# ---------- import later
# from mpi4py import MPI
# from cryspy.job.ctrl_ext import Ctrl_ext
# from cryspy.job.ctrl_job import Ctrl_job
# from cryspy.interface import select_code

def main():
    # ---------- argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-b', '--backup', help='backup data', action='store_true')
    parser.add_argument('-c', '--clean', help='clean data', action='store_true')
    parser.add_argument('-g', '--debug', help='debug', action='store_true')
    parser.add_argument('-n', '--noprint', help='not printing to the console', action='store_true')
    parser.add_argument('-p', '--parallel', help='with MPI', action='store_true')
    args = parser.parse_args()

    # ########## MPI start
    # ---------- MPI
    if args.parallel:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        mpi_rank = comm.Get_rank()
        mpi_size = comm.Get_size()
    else:
        # ------ normal run
        comm = None
        mpi_rank = 0
        mpi_size = 1

    # ---------- logger
    set_logger(
        noprint=args.noprint,
        debug=args.debug,
        logfile='log_cryspy',
        errfile='err_cryspy',
        debugfile='debug_cryspy',
    )
    logger = getLogger('cryspy')

    # ---------- backup option
    if args.backup:
        if mpi_rank == 0:
            backup_cryspy()
        raise SystemExit()

    # ---------- clean option
    if args.clean:
        if mpi_rank == 0:
            clean_cryspy()
        raise SystemExit()

    # ---------- lock
    if os.path.isfile('lock_cryspy'):
        if mpi_rank == 0:
            logger.error('lock_cryspy file exists')
        raise SystemExit(1)
    else:
        if mpi_size > 1:
            comm.barrier()
        if mpi_rank == 0:
            with open('lock_cryspy', 'w'):
                pass    # create vacant file
        else:
            pass

    # ---------- initialize
    if not os.path.isfile('cryspy.stat'):
        cryspy_init.initialize(comm, mpi_rank, mpi_size)
        if mpi_rank == 0:
            os.remove('lock_cryspy')
        raise SystemExit()
    # ---------- restart
    else:
        # only stat and init_struc_data in rank0 are important
        rin, init_struc_data = cryspy_restart.restart(comm, mpi_rank, mpi_size)
    # ########## MPI end

    if mpi_rank == 0:
        # ---------- check point 1
        if rin.stop_chkpt == 1:
            logger.info('Stop at check point 1')
            os.remove('lock_cryspy')
            raise SystemExit()

        if not rin.calc_code == 'ext':
            # ---------- check calc files in ./calc_in
            from cryspy.interface import select_code
            select_code.check_calc_files(rin)
            # ---------- mkdir work/fin
            os.makedirs('work/fin', exist_ok=True)

        # ---------- Perform structure optimization externally
        if rin.calc_code == 'ext':
            # ------ instantiate Ctrl_job class
            from cryspy.job.ctrl_ext import Ctrl_ext
            jobs = Ctrl_ext(rin, init_struc_data)
        # ---------- internally
        else:
            # ---------- instantiate Ctrl_job class
            from cryspy.job.ctrl_job import Ctrl_job
            jobs = Ctrl_job(rin, init_struc_data)

        # ---------- check job status
        jobs.check_job()

        # ---------- handle job
        jobs.handle_job()

        # ---------- recheck for skip and done
        if not rin.calc_code == 'ext':
            if jobs.id_queueing:
                cnt_recheck = 0
                while jobs.recheck:
                    cnt_recheck += 1
                    jobs.recheck = False    # True --> False
                    logger.info(f'\n\nrecheck {cnt_recheck}\n')
                    jobs.check_job()
                    jobs.handle_job()

        # ---------- next selection or generation
        if not (jobs.id_queueing or jobs.id_running):
            # ---------- next selection or generation
            if rin.algo in ['BO', 'LAQA', 'EA', 'EA-vc']:
                jobs.next_sg(args.noprint)
            # ---------- for RS
            else:
                logger.info('\nDone all structures!')

        # ---------- unlock
        os.remove('lock_cryspy')


if __name__ == '__main__':
    # ---------- main
    main()



from logging import getLogger
import os
import shutil


logger = getLogger('cryspy')

def next_stage_ase(rin, stage, work_path):
    # ---------- skip_flag
    skip_flag = False

    # ---------- rename ASE files at the current stage
    ase_files = ['POSCAR', 'CONTCAR', 'log.tote']
    for file in ase_files:
        if not os.path.isfile(work_path + file):
            logger.error('Not found ' + work_path + file)
            raise SystemExit(1)
        os.rename(work_path + file, work_path + f'stage{stage}_' + file)

    # ---------- cp CONTCAR POSCAR
    shutil.copyfile(work_path + f'stage{stage}_CONTCAR', work_path + 'POSCAR')

    # ---------- copy the input file from ./calc_in for the next stage
    finfile = './calc_in/' +rin.ase_python + f'_{stage+1}'
    shutil.copyfile(finfile, work_path+rin.ase_python)

    # ---------- return
    return skip_flag


def next_struc_ase(rin, structure, cid, work_path):
    # ---------- copy files
    calc_inputs = [rin.ase_python]
    for f in calc_inputs:
        ff = f+'_1' if f == rin.ase_python else f
        if not os.path.isfile('./calc_in/' + ff):
            logger.error('Could not find ./calc_in/' + ff)
            raise SystemExit(1)
        # ------ e.g. cp ./calc_in/INCAR_1 work/1/INCAR
        shutil.copyfile('./calc_in/'+ff, work_path+f)

    # ---------- generate POSCAR
    structure.to(fmt='poscar', filename=work_path+'POSCAR')
    if not os.path.isfile(work_path+'POSCAR'):
        logger.error(f'Could not find {work_path}POSCAR')

    # ---------- Change the title of POSCAR
    with open(work_path+'POSCAR', 'r') as f:
        lines = f.readlines()
    lines[0] = f'ID_{cid}\n'
    with open(work_path+'POSCAR', 'w') as f:
        for line in lines:
            f.write(line)



from logging import getLogger


logger = getLogger('cryspy')


def diff_in(rin, pin):
    '''
    Args:
        rin: ReadInput
        pin: previous ReadInput
    '''

    # ---------- compare current and previous input
    if rin == pin:
        return

    # ---------- immutables
    if rin.algo not in ['EA', 'EA-vc']:
        immutables = (
            'algo', 'calc_code', 'atype', 'natot', 'nat', 'nmol'
            'energy_step_flag', 'struc_step_flag', 'force_step_flag', 'stress_step_flag',
            'dscrpt', 'fp_rmin', 'fp_rmax', 'fp_npoints', 'fp_sigma',
        )
    else:    # EA, EA-vc
        immutables = (
            'algo', 'calc_code', 'atype', 'tot_struc',
            'energy_step_flag', 'struc_step_flag', 'force_step_flag', 'stress_step_flag',
        )

    # ----------
    for key in rin.__annotations__.keys():
        if getattr(rin, key) != getattr(pin, key):
            if key in immutables:
                raise ValueError(f'Do not change {key}')
            else:
                logger.info(f'Changed {key} from {getattr(pin, key)} to {getattr(rin, key)}')


'''
Random structure generation w/o pyxtal
'''

import json
from logging import getLogger
import math
import os
import random
import subprocess

import numpy as np
from pymatgen.core import Structure

from ...util.struc_util import check_distance


logger = getLogger('cryspy')


def gen_wo_spg(
        nstruc,
        atype,
        nat,
        mindist,
        spgnum,
        minlen,
        maxlen,
        dangle,
        symprec=0.01,
        maxcnt=50,
        id_offset=0,
        vol_mu=None,
        vol_sigma=None,
        vc=False,
        ll_nat=None,
        ul_nat=None,
    ):
    '''
    Generate random structures without space group information

    # ---------- args
    nstruc (int): number of structures to be generated
    atype (tuple): atom types, e.g. ('Na', 'Cl')
    nat (tuple): number of atoms, e.g. (8, 8)
    mindist (): minumum interatomic distance, e.g. ((2.0, 1.5), (1.5, 2.0))
    spgnum (str, int, or tuple): space group number 'all', 0, or tuple of space group numbers
    minlen (float): minimum length of lattice vector
    maxlen (float): maximum length of lattice vector
    dangle (float): maximum deviation of lattice angle
    symprec (float): tolerance to find space group
    maxcnt (int): maximum trial to generate a structure
    id_offset (int): structure ID starts from id_offset
                        e.g. nstruc = 3, id_offset = 10
                            you obtain ID 10, ID 11, ID 12
    vol_mu (float): mean of volume scaling
    vol_sigma (float): standard deviation of volume scaling
    vc (bool): variable composition. it needs ll_nat and ul_nat
    ll_nat (tuple): lower limit of number of atoms, e.g. (1, 1)
    ul_nat (tuple): upper limit of number of atoms, e.g. (8, 8)

    # ---------- return
    init_struc_data (dict): {ID: pymatgen structure data}
    '''

    # ---------- initialize
    init_struc_data = {}

    # ---------- generate structures
    while len(init_struc_data) < nstruc:
        # ------ vc
        if not vc:
            numIons = nat
        else:    # variable composition
            numIons = [random.randint(l, u) for l, u in zip(ll_nat, ul_nat)]
            numIons = tuple(numIons)
        atomlist = _get_atomlist(atype, numIons)
        # ------ get spg, a, b, c, alpha, beta, gamma
        spg, a, b, c, alpha, beta, gamma = _gen_lattice(spgnum, minlen, maxlen, dangle)
        # ------ get a1, a2, a3
        a1, a2, a3 = _calc_latvec(a, b, c, alpha, beta, gamma)
        # ------ get structure
        tmp_struc = _gen_struc_wo_spg(atype, numIons, atomlist, a1, a2, a3, mindist, maxcnt)
        if tmp_struc is not None:    # success of generation
            # ------ scale volume
            if vol_mu is not None:
                vol = random.gauss(mu=vol_mu, sigma=vol_sigma)
                tmp_struc.scale_lattice(volume=vol)
                success, mindist_ij, dist = check_distance(tmp_struc, atype, mindist)
                if not success:
                    type0 = atype[mindist_ij[0]]
                    type1 = atype[mindist_ij[1]]
                    logger.warning(f'mindist: {type0} - {type1}, {dist}. retry.')
                    continue    # failure
            # ------ check actual space group using pymatgen
            try:
                spg_sym, spg_num = tmp_struc.get_space_group_info(symprec=symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
            # ------ register the structure in pymatgen format
            cid = len(init_struc_data) + id_offset
            init_struc_data[cid] = tmp_struc
            logger.info(f'Structure ID {cid:>6}: {numIons}'
                    f' Space group: {spg_num:>3} {spg_sym}')

    # ---------- return
    return init_struc_data


def gen_with_find_wy(
        nstruc,
        atype,
        nat,
        mindist,
        spgnum,
        minlen,
        maxlen,
        dangle,
        symprec=0.01,
        maxcnt=50,
        id_offset=0,
        vol_mu=None,
        vol_sigma=None,
        fwpath='find_wy',
        mpi_rank=0,
        vc=False,
        ll_nat=None,
        ul_nat=None,
    ):
    '''
    Generate random structures with space gruop information
    using find_wy program

    # ---------- args
    nstruc (int): number of generated structures
    atype (tuple): atom types, e.g. ('Na', 'Cl')
    nat (tuple): number of atoms, e.g. (8, 8)
    mindist (): minimum interatomic distance e.g. ((2.0, 1.5), (1.5, 2.0))
    spgnum (str, int, or tuple): space group number 'all', 0, or tuple of space group numbers
    minlen (float): minimum length of lattice vector
    maxlen (float): maximum length of lattice vector
    dangle (float): maximum deviation of lattice angle
    symprec (float): tolerance to find space group
    maxcnt (int): maximum trial to generate a structure
    id_offset (int): structure ID starts from id_offset
                        e.g. nstruc = 3, id_offset = 10
                            you obtain ID 10, ID 11, ID 12
    vol_mu (float): mean of volume scaling
    vol_sigma (float): standard deviation of volume scaling
    fwpath (str): specify a path for a executable file of find_wy program
    mpi_rank (int): rank of MPI process
    vc (bool): variable composition. it needs ll_nat and ul_nat
    ll_nat (tuple): lower limit of number of atoms, e.g. (1, 1)
    ul_nat (tuple): upper limit of number of atoms, e.g. (8, 8)

    # ---------- return
    init_struc_data (dict): {ID: pymatgen Structre}
    '''

    # ---------- initialize
    init_struc_data = {}

    # ---------- cd tmp_gen_struc
    os.makedirs(f'tmp_gen_struc/rank_{mpi_rank}', exist_ok=True)
    os.chdir(f'tmp_gen_struc/rank_{mpi_rank}')

    # ---------- generate structures
    while len(init_struc_data) < nstruc:
        # ------ vc
        if not vc:
            numIons = nat
        else:    # variable composition
            numIons = [random.randint(l, u) for l, u in zip(ll_nat, ul_nat)]
            numIons = tuple(numIons)
        # ------ get spg, a, b, c, alpha, beta, gamma
        spg, a, b, c, alpha, beta, gamma = _gen_lattice(spgnum, minlen, maxlen, dangle)
        # ------ get cosa, cosb, and cosc
        cosa, cosb, cosg = _calc_cos(alpha, beta, gamma)
        # ------ write an input file for find_wy
        _fw_input(atype, numIons, spg, a, b, c, cosa, cosb, cosg)
        # ------ loop for same fw_input
        cnt = 0
        while cnt <= maxcnt:
            # -- run find_wy
            with open('log_find_wy', 'w') as f:
                subprocess.call([fwpath, 'input'], stdout=f, stderr=f)
            # -- generate a structure using POS_WY_SKEL_ALL.json
            if not os.path.isfile('POS_WY_SKEL_ALL.json'):
                wyflag = False
                break
            wyflag, tmp_struc = _gen_struc_with_spg(atype, mindist, maxcnt)
            if wyflag is False:    # Failure
                os.remove('POS_WY_SKEL_ALL.json')
                cnt += 1
                continue
            else:    # Success
                _rm_files()    # rm input POS_WY_SKEL_ALL.json
                break          # break fw_input loop
        if wyflag is False:
            # -- maximum trial or no POS_WY_SKEL_ALL.json file
            _rm_files()    # clean
            continue       # to new fw_input
        # ------ scale volume
        if vol_mu is not None:
            vol = random.gauss(mu=vol_mu, sigma=vol_sigma)
            tmp_struc.scale_lattice(volume=vol)
            success, mindist_ij, dist = check_distance(tmp_struc, atype, mindist)
            if not success:
                type0 = atype[mindist_ij[0]]
                type1 = atype[mindist_ij[1]]
                logger.warning(f'mindist: {type0} - {type1}, {dist}. retry.')
                continue    # failure
        # ------ check actual space group using pymatgen
        try:
            spg_sym, spg_num = tmp_struc.get_space_group_info(
                symprec=symprec)
        except TypeError:
            spg_num = 0
            spg_sym = None
        # ------ tmp_struc --> init_struc_data
        cid = len(init_struc_data) + id_offset
        init_struc_data[cid] = tmp_struc
        logger.info(f'Structure ID {cid:>6}: {numIons}'
                f' Space group: {spg:>3} --> {spg_num:>3} {spg_sym}')
        # ------ clean
        _rm_files()

    # ---------- go back to ..
    os.chdir('../../')

    # ---------- return
    return init_struc_data


def _get_atomlist(atype, numIons):
    '''
    e.g. Na2Cl2
        atomlist = ['Na', 'Na', 'Cl', 'Cl']
    '''
    atomlist = []
    for i in range(len(atype)):
        atomlist += [atype[i]]*numIons[i]
    return atomlist

def _gen_lattice(spgnum, minlen, maxlen, dangle):
    # ---------- for spgnum = 0: no space group
    if spgnum == 0:
        crystal_systems = ['Triclinic',
                            'Monoclinic',
                            'Orthorhombic',
                            'Tetragonal',
                            'Rhombohedral',
                            'Hexagonal',
                            'Cubic']
        spg = 0
        csys = random.choice(crystal_systems)
    # ---------- for spgnum 1--230
    else:
        # ------ spgnum --> spg
        if spgnum == 'all':
            spg = random.randint(1, 230)
        else:
            spg = random.choice(spgnum)
        if 1 <= spg <= 2:
            csys = 'Triclinic'
        elif 3 <= spg <= 15:
            csys = 'Monoclinic'
        elif 16 <= spg <= 74:
            csys = 'Orthorhombic'
        elif 75 <= spg <= 142:
            csys = 'Tetragonal'
        elif 143 <= spg <= 167:
            # trigonal includes rhombohedral in find_wy
            csys = 'Trigonal'
        elif 168 <= spg <= 194:
            csys = 'Hexagonal'
        elif 195 <= spg <= 230:
            csys = 'Cubic'
        else:
            logger.error('spg is wrong')
            raise SystemExit(1)
    # ---------- generate lattice constants a, b, c, alpha, beta, gamma
    if csys == 'Triclinic':
        t1 = random.uniform(minlen, maxlen)
        t2 = random.uniform(minlen, maxlen)
        t3 = random.uniform(minlen, maxlen)
        t = [t1, t2, t3]
        t.sort()
        a, b, c = t
        r = random.random()
        if r < 0.5:    # Type I
            alpha = 90.0 - random.uniform(0, dangle)
            beta  = 90.0 - random.uniform(0, dangle)
            gamma = 90.0 - random.uniform(0, dangle)
        else:    # Type II
            alpha = 90.0 + random.uniform(0, dangle)
            beta  = 90.0 + random.uniform(0, dangle)
            gamma = 90.0 + random.uniform(0, dangle)
    elif csys == 'Monoclinic':
        a = random.uniform(minlen, maxlen)
        b = random.uniform(minlen, maxlen)
        c = random.uniform(minlen, maxlen)
        if a > c:
            a, c = c, a
        alpha = gamma = 90.0
        beta = 90.0 + random.uniform(0, dangle)
    elif csys == 'Orthorhombic':
        t1 = random.uniform(minlen, maxlen)
        t2 = random.uniform(minlen, maxlen)
        t3 = random.uniform(minlen, maxlen)
        t = [t1, t2, t3]
        t.sort()
        a, b, c = t
        alpha = beta = gamma = 90.0
    elif csys == 'Tetragonal':
        a = b = random.uniform(minlen, maxlen)
        c = random.uniform(minlen, maxlen)
        alpha = beta = gamma = 90.0
    elif csys == 'Trigonal':
        a = b = random.uniform(minlen, maxlen)
        c = random.uniform(minlen, maxlen)
        alpha = beta = 90.0
        gamma = 120.0
    elif csys == 'Rhombohedral':
        a = b = c = random.uniform(minlen, maxlen)
        alpha = beta = gamma = 90 + random.uniform(-dangle,
                                                    dangle)
    elif csys == 'Hexagonal':
        a = b = random.uniform(minlen, maxlen)
        c = random.uniform(minlen, maxlen)
        alpha = beta = 90.0
        gamma = 120.0
    elif csys == 'Cubic':
        a = b = c = random.uniform(minlen, maxlen)
        alpha = beta = gamma = 90.0

    # ---------- return
    return spg, a, b, c, alpha, beta, gamma


def _calc_latvec(a, b, c, alpha, beta, gamma):
    # ---------- degree to radian
    alpha_rad = math.radians(alpha)
    beta_rad = math.radians(beta)
    gamma_rad = math.radians(gamma)
    # ---------- calculate components
    bx = b*math.cos(gamma_rad)
    by = b*math.sin(gamma_rad)
    cx = c*math.cos(beta_rad)
    cy = (c*math.cos(alpha_rad)
            - cx*math.cos(gamma_rad))/math.sin(gamma_rad)
    cz = math.sqrt(c*c - cx*cx - cy*cy)
    # ---------- lattice vector as list
    a1 = [a, 0.0, 0.0]
    a2 = [bx, by, 0.0]
    a3 = [cx, cy, cz]
    # ---------- return
    return a1, a2, a3


def _calc_cos(alpha, beta, gamma):
    # ---------- degree to radian
    a_rad = math.radians(alpha)
    b_rad = math.radians(beta)
    g_rad = math.radians(gamma)
    cosa = math.cos(a_rad)
    cosb = math.cos(b_rad)
    cosg = math.cos(g_rad)
    # ---------- return
    return cosa, cosb, cosg


def _gen_struc_wo_spg(atype, numIons, atomlist, a1, a2, a3, mindist, maxcnt=50):
    '''
    Success --> return structure data in pymatgen format
    Failure --> return None
    '''
    # ---------- initialize
    cnt = 0
    incoord = []
    natot = sum(numIons)    # do not use rin.natot because of vc
    # ---------- generate internal coordinates
    while len(incoord) < natot:
        tmp_coord = np.random.rand(3)
        incoord.append(tmp_coord)
        tmp_struc = Structure([a1, a2, a3],
                                atomlist[:len(incoord)],
                                incoord)
        success, mindist_ij, dist = check_distance(tmp_struc, atype, mindist)
        if not success:
            type0 = atype[mindist_ij[0]]
            type1 = atype[mindist_ij[1]]
            logger.warning(f'mindist: {type0} - {type1}, {dist}. retry.')
            incoord.pop(-1)    # cancel
            cnt += 1
            if maxcnt < cnt:
                return None
    return tmp_struc


def _fw_input(atype, numIons, spg, a, b, c, cosa, cosb, cosg):
    with open('input', 'w') as f:
        f.write(f'nspecies {len(atype)}\n')
        f.write('species_name')
        for aa in atype:
            f.write(f'  {aa}')
        f.write('\n')
        f.write('species_num')
        for i in numIons:
            f.write(f'  {i}')
        f.write('\n')
        f.write(f'spacegroup  {spg}\n')
        f.write('originchoice  1\n')
        f.write('\n')
        f.write(f'a  {a}\n')
        f.write(f'b  {b}\n')
        f.write(f'c  {c}\n')
        f.write(f'cosa  {cosa}\n')
        f.write(f'cosb  {cosb}\n')
        f.write(f'cosc  {cosg}\n')
        f.write('\n')
        # f.write('selectone true\n')
        f.write('randomseed auto\n')


def _gen_struc_with_spg(atype, mindist, maxcnt):
    '''
    Success --> return True, structure data
    Failure --> return False, _
    '''
    # ---------- load POS_WY_SKEL_ALL.json
    with open('POS_WY_SKEL_ALL.json', 'r') as f:
        wydata = json.load(f)
    # ---------- generate structure
    plat = wydata['primitivevector']
    clat = wydata['conventionalvector']
    n_uniq, wydata_eq_atom = _get_wydata_eq_atom(wydata)
    eq_atomnames = {}
    eq_positions = {}
    # ---------- equivalent atom loop
    for key, value in sorted(n_uniq.items(), key=lambda x: x[1]):
        # ------ distribute eq atoms. first, special (num_uniqvar = 0),
        #            then, others
        cnt = 0
        while True:
            eq_atomnames[key], eq_positions[key] = _gen_eq_atoms(
                wydata_eq_atom[key])
            # -- sort in original order
            atomnames = []
            positions = []
            for key_a, value_a in sorted(eq_atomnames.items()):
                atomnames += eq_atomnames[key_a]
                positions += eq_positions[key_a]
            # -- Cartesian coordinate; use clat (not plat)
            cart = []
            for p in positions:
                v = np.zeros(3)
                for i in range(3):
                    a = np.array(clat[i])
                    v += p[i] * a
                cart.append(v)
            # -- check minimum distance
            spgstruc = Structure(plat, atomnames, cart,
                                    coords_are_cartesian=True)
            success, mindist_ij, dist = check_distance(spgstruc, atype, mindist)
            if not success:
                type0 = atype[mindist_ij[0]]
                type1 = atype[mindist_ij[1]]
                logger.warning(f'mindist: {type0} - {type1}, {dist}. retry.')
                # failure
                # num_uniqvar = 0 --> value == 0
                cnt = maxcnt + 1 if value == 0 else cnt + 1
                if maxcnt < cnt:
                    return False, spgstruc    # spgstruc is dummy
            else:
                break    # break while loop --> next eq atoms
    return True, spgstruc

def _get_wydata_eq_atom(wydata):
    i = 0    # count eq_atom, not atom
    n_uniq = {}    # num_uniqvar each eq_atom
    wydata_eq_atom = {}    # wydata each eq_atom
    for specie in wydata['atoms']:
        for wydata2 in specie:    # equivalent atom loop
            n_uniq[i] = wydata2[0]['num_uniqvar']
            wydata_eq_atom[i] = wydata2
            i += 1
    return n_uniq, wydata_eq_atom

def _gen_eq_atoms(wydata2):
    eq_atomnames = []
    eq_positions = []
    rval = np.random.random_sample(3)
    for each in wydata2:
        pos = []
        for ch in each['xyzch']:
            if ch == '-2x':
                pos.append(-2.0 * rval[0])
            elif ch == '-x+y':
                pos.append(-rval[0] + rval[1])
            elif ch == '-z':
                pos.append(-rval[2])
            elif ch == '-y':
                pos.append(-rval[1])
            elif ch == '-x':
                pos.append(-rval[0])
            elif ch == '0':
                pos.append(0.0)
            elif ch == 'x':
                pos.append(rval[0])
            elif ch == 'y':
                pos.append(rval[1])
            elif ch == 'z':
                pos.append(rval[2])
            elif ch == 'x-y':
                pos.append(rval[0] - rval[1])
            elif ch == '2x':
                pos.append(2.0 * rval[0])
            else:
                logger.error('unknown ch in conversion in gen_wycoord')
                raise SystemExit(1)
        pos = np.array(pos)
        eq_positions.append(pos + each['add'])
        eq_atomnames.append(each['name'])
    return eq_atomnames, eq_positions

def _rm_files(files=['input', 'POS_WY_SKEL_ALL.json']):
    for rfile in files:
        if os.path.isfile(rfile):
            os.remove(rfile)



'''
Calculation files in OpenMX
written by H. Sawahata 2020/03/09
info at hikaruri.jp
'''

from logging import getLogger
import os


logger = getLogger('cryspy')


def check_input_OMX(rin):
    # ---------- prepare rin.jobfile, rin.OMX_infile
    calc_inputs = [rin.jobfile, rin.OMX_infile]

    # ------ check required files
    for f in calc_inputs:
        if f == rin.OMX_infile:
            finfiles = [rin.OMX_infile + f'_{i}' for i in range(
                1, rin.nstage+1)]
            for ff in finfiles:
                if not os.path.isfile('./calc_in/' + ff):
                    logger.error('Could not find ./calc_in/' + ff)
                    raise SystemExit(1)
        else:
            if not os.path.isfile('./calc_in/' + f):
                logger.error('Could not find ./calc_in/' + f)
                raise SystemExit(1)



from logging import getLogger
import os
import shutil

import numpy as np

from ..interface import select_code
from ..IO import io_stat, pkl_data
from ..IO.out_results import out_rslt, out_hdist
from ..util.utility import backup_cryspy
from ..util.struc_util import out_poscar, out_cif

# ---------- import later
#from ..BO.select_descriptor import select_descriptor
#from ..BO import bo_next_select
#from ..EA import ea_next_gen
#from ..EA.calc_ef import calc_ef


logger = getLogger('cryspy')


class Ctrl_ext:

    def __init__(self, rin, init_struc_data):
        self.rin = rin
        self.init_struc_data = init_struc_data
        self.opt_struc_data = pkl_data.load_opt_struc()
        self.rslt_data = pkl_data.load_rslt()
        self.id_queueing = pkl_data.load_id_queueing()
        self.id_running = pkl_data.load_id_running()
        self.recheck = False
        self.logic_next = False
        # ---------- for each algorithm
        if rin.algo == 'RS':
            pass
        elif rin.algo == 'BO':
            self.n_selection = pkl_data.load_n_selection()
            self.id_select_hist = pkl_data.load_id_select_hist()
            self.init_dscrpt_data = pkl_data.load_init_dscrpt_data()
            self.opt_dscrpt_data = pkl_data.load_opt_dscrpt_data()
            self.bo_mean = pkl_data.load_bo_mean()
            self.bo_var = pkl_data.load_bo_var()
            self.bo_score = pkl_data.load_bo_score()
        elif rin.rin.algo in ['EA', 'EA-vc']:
            self.gen = pkl_data.load_gen()
            if rin.algo == 'EA-vc':
                self.nat_data = pkl_data.load_nat_data()
                self.ratio_data = pkl_data.load_ratio_data()
                self.hdist_data = pkl_data.load_hdist_data()
            # do not have to load ea_data here.
            # ea_data is used only in ea_next_gen.py

    def check_job(self):
        # ---------- option: recalc
        #if self.rin.recalc is not None:
        #    self.set_recalc()
        # ---------- check job status
        try:
            with open('ext/stat_job') as fstat:
                jstat = fstat.readline()    # submitted or done or ...
            if jstat[0:3] == 'sub':
                self.job_stat = 'submitted'
            elif jstat[0:4] == 'done':
                self.job_stat = 'done'
            elif jstat[0:3] == 'out':
                self.job_stat = 'out'
            elif jstat[0:2] == 'no':
                self.job_stat = 'no queue'
            else:
                self.job_stat = 'else'
        except IOError:
            self.job_stat = 'no_file'

    def handle_job(self):
        logger.info('# ---------- job status')
        if self.job_stat == 'submitted':
            logger.info('still queueing or running')
        elif self.job_stat == 'done':
            logger.info('collect data')
            self.ctrl_done()
        elif self.job_stat == 'out':
            logger.info('write queueing structure data in ext/queue/')
            self.out_queue()
        elif self.job_stat == 'no queue':
            pass
        elif self.job_stat == 'else':
            logger.error('Wrong job_stat')
            raise SystemExit(1)
        elif self.job_stat == 'no_file':
            logger.error('Wrong job_stat')
            raise SystemExit(1)
        else:
            logger.error('Unexpected error')
            raise SystemExit(1)

    def out_queue(self):
        # ---------- out cifs
        os.makedirs('ext/queue', exist_ok=False)    # if dir exists, error --> stop
        for cid in self.id_queueing:
            self.init_struc_data[cid].to(fmt='cif', filename=f'ext/queue/{cid}.cif')
        # ---------- queue --> running
        self.id_running = self.id_queueing[:]
        self.id_queueing = []
        stat = io_stat.stat_read()
        io_stat.set_id(stat, 'id_queueing', self.id_queueing)
        io_stat.write_stat(stat)
        self.save_id_data()
        # ---------- stat_job
        with open('ext/stat_job', 'w') as fstat:
            fstat.write('submitted\n')
        # ---------- mkdir calc_data
        os.makedirs('ext/calc_data', exist_ok=True)

    def ctrl_done(self):
        # ---------- collect result
        self.ctrl_collect()

    def ctrl_collect(self):
        # ---------- each algo
        if self.rin.algo == 'RS':
            self.ctrl_collect_rs()
        elif self.rin.algo == 'BO':
            self.ctrl_collect_bo()
        elif self.rin.algo in ['EA', 'EA-vc']:
            self.ctrl_collect_ea()
        else:
            logger.error('Error, algo')
            raise SystemExit(1)
        # ---------- calc_data --> old_calc_data
        if os.path.isdir('ext/old_calc_data'):
            shutil.rmtree('ext/old_calc_data')
        shutil.move('ext/calc_data', 'ext/old_calc_data')
        # ---------- queue --> old_queue
        if os.path.isdir('ext/old_queue'):
            shutil.rmtree('ext/old_queue')
        shutil.move('ext/queue', 'ext/old_queue')
        # ---------- stat_job
        with open('ext/stat_job', 'w') as fstat:
            fstat.write('no queue\n')
        # ---------- update status
        self.id_running = []
        self.save_id_data()

    def ctrl_collect_rs(self):
        # ---------- get opt data
        ext_opt_struc_data, ext_energy_data = select_code.collect(self.rin, 'dummy', 'dummy', 'dummy')
        # ---------- register opt_struc
        ext_magmom, ext_check_opt, ext_spg_sym, ext_spg_num, \
            ext_spg_sym_opt, ext_spg_num_opt = self.regist_opt(ext_opt_struc_data)
        # ---------- save rslt
        for cid, opt_struc in ext_opt_struc_data.items():
            self.rslt_data.loc[cid] = [ext_spg_num[cid], ext_spg_sym[cid],
                                       ext_spg_num_opt[cid], ext_spg_sym_opt[cid],
                                       ext_energy_data[cid], ext_magmom[cid], ext_check_opt[cid]]
        pkl_data.save_rslt(self.rslt_data)
        out_rslt(self.rslt_data)

    def ctrl_collect_bo(self):
        # ---------- get opt data
        ext_opt_struc_data, ext_energy_data = select_code.collect(self.rin, 'dummy', 'dummy', 'dummy')
        # ---------- register opt_struc
        ext_magmom, ext_check_opt, ext_spg_sym, ext_spg_num, \
            ext_spg_sym_opt, ext_spg_num_opt = self.regist_opt(ext_opt_struc_data)
        # ---------- save rslt
        for cid, opt_struc in ext_opt_struc_data.items():
            self.rslt_data.loc[cid] = [self.n_selection,
                                       ext_spg_num[cid], ext_spg_sym[cid],
                                       ext_spg_num_opt[cid], ext_spg_sym_opt[cid],
                                       ext_energy_data[cid], ext_magmom[cid], ext_check_opt[cid]]
        pkl_data.save_rslt(self.rslt_data)
        out_rslt(self.rslt_data)
        # ---------- descriptor
        for cid, opt_struc in ext_opt_struc_data.items():
            if opt_struc is not None:
                # ------ calc descriptor for opt structure
                from ..BO.select_descriptor import select_descriptor
                tmp_dscrpt = select_descriptor(self.rin, {cid: opt_struc})
                # ------ update descriptors
                self.opt_dscrpt_data.update(tmp_dscrpt)
                # ---------- error
            else:
                # ------ update descriptors and non_error_id
                self.opt_dscrpt_data[cid] = None
        # ---------- save bo_data
        self.save_data()

    def ctrl_collect_ea(self):
        # ---------- get opt data
        ext_opt_struc_data, ext_energy_data = select_code.collect(self.rin, 'dummy', 'dummy', 'dummy')
        # ---------- register opt_struc
        ext_magmom, ext_check_opt, ext_spg_sym, ext_spg_num, \
            ext_spg_sym_opt, ext_spg_num_opt = self.regist_opt(ext_opt_struc_data)
        # ---------- save rslt
        for cid, opt_struc in ext_opt_struc_data.items():
            if not self.rin.algo == 'EA-vc':
                self.rslt_data.loc[cid] = [self.gen,
                                           ext_spg_num[cid], ext_spg_sym[cid],
                                           ext_spg_num_opt[cid], ext_spg_sym_opt[cid],
                                           ext_energy_data[cid], ext_magmom[cid], ext_check_opt[cid]]
            else:    # for EA-vc
                from ..EA.calc_ef import calc_ef
                ef = calc_ef(ext_energy_data[cid], self.ratio_data[cid],
                             self.rin.end_point)
                self.rslt_data.loc[cid] = [self.gen,
                                           ext_spg_num[cid], ext_spg_sym[cid],
                                           ext_spg_num_opt[cid], ext_spg_sym_opt[cid],
                                           ext_energy_data[cid], ef, ext_magmom[cid], ext_check_opt[cid]]
        pkl_data.save_rslt(self.rslt_data)
        out_rslt(self.rslt_data)

    def regist_opt(self, ext_opt_struc_data):
        '''
        Common part in ctrl_collect_*
        '''
        # ---------- space group etc.
        ext_magmom = {}
        ext_check_opt = {}
        ext_spg_sym = {}
        ext_spg_num = {}
        ext_spg_sym_opt = {}
        ext_spg_num_opt = {}
        for cid, opt_struc in ext_opt_struc_data.items():
            ext_magmom[cid] = np.nan
            ext_check_opt[cid] = 'no_file'
            # ------ get initial spg info
            try:
                ext_spg_sym[cid], ext_spg_num[cid] = self.init_struc_data[
                    cid].get_space_group_info(symprec=self.rin.symprec)
            except TypeError:
                ext_spg_num[cid] = 0
                ext_spg_sym[cid] = None
            # ------ success opt
            if ext_opt_struc_data[cid] is not None:
                # -- get opt spg info
                try:
                    ext_spg_sym_opt[cid], ext_spg_num_opt[cid] = opt_struc.get_space_group_info(
                        symprec=self.rin.symprec)
                except TypeError:
                    ext_spg_num_opt[cid] = 0
                    ext_spg_sym_opt[cid] = None
                # -- out opt_struc
                out_poscar({cid:opt_struc}, './data/opt_POSCARS')
                try:
                    out_cif(opt_struc, cid, './data/',
                            './data/opt_CIFS.cif', self.rin.symprec)
                except TypeError:
                    logger.info('failed to write opt_CIF')
            # ------ error
            else:
                ext_spg_num_opt[cid] = 0
                ext_spg_sym_opt[cid] = None
        # ---------- save opt_struc_data
        self.opt_struc_data.update(ext_opt_struc_data)
        pkl_data.save_opt_struc(self.opt_struc_data)
        # ---------- return
        return ext_magmom, ext_check_opt, ext_spg_sym, ext_spg_num, ext_spg_sym_opt, ext_spg_num_opt

    def next_sg(self, noprint=False):
        '''
        next selection or generation
        '''
        if self.rin.algo == 'BO':
            self.next_select_BO(noprint)
        if self.rin.algo in ['EA', 'EA-vc']:
            self.next_gen_EA()

    def next_select_BO(self, noprint=False):
        # ---------- log
        logger.info(f'Done selection {self.n_selection}')
        # ---------- done all structures
        if len(self.rslt_data) == self.rin.tot_struc:
            logger.info('\nDone all structures!')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- check point 3
        if self.rin.stop_chkpt == 3:
            logger.info('\nStop at check point 3: BO is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- max_select_bo
        if 0 < self.rin.max_select_bo <= self.n_selection:
            logger.info(f'\nReached max_select_bo: {self.rin.max_select_bo}')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- BO
        bo_data = (self.init_dscrpt_data, self.opt_dscrpt_data,
                   self.bo_mean, self.bo_var, self.bo_score)
        bo_id_data = (self.n_selection, self.id_queueing,
                      self.id_running, self.id_select_hist)
        from ..BO import bo_next_select
        bo_next_select.next_select(self.rin, self.rslt_data,
                                   bo_id_data, bo_data, noprint)

    def next_gen_EA(self):
        # ---------- log
        logger.info(f'Done generation {self.gen}')
        # ---------- check point 3
        if self.rin.stop_chkpt == 3:
            logger.info('\nStop at check point 3: EA is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- maxgen_ea
        if 0 < self.rin.maxgen_ea <= self.gen:
            if self.rin.algo == 'EA-vc':
                # ------ when gen reaches maxgen_ea,  next generation is not created
                #        so convex hull is calculated here
                #        update only convex hull and hdist, not elite_struc and elite_fitness
                c_rslt = self.rslt_data[self.rslt_data['Gen'] == self.gen]
                c_ids = c_rslt.index.values    # current IDs [array]
                ef_all = self.rslt_data['Ef_eV_atom'].to_dict()    # formation energy of all structures
                from ..EA.calc_hull import calc_convex_hull_2d
                hdist = calc_convex_hull_2d(self.rin, self.ratio_data, ef_all, c_ids, self.gen)
                out_hdist(self.gen, hdist, self.ratio_data)
                self.hdist_data[self.gen] = hdist
                pkl_data.save_hdist_data(self.hdist_data)
            logger.info(f'\nReached maxgen_ea: {self.rin.maxgen_ea}')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- EA
        backup_cryspy()
        ea_id_data = (self.gen, self.id_queueing, self.id_running)
        if self.rin.algo == 'EA-vc':
            ea_vc_data = (self.nat_data, self.ratio_data, self.hdist_data)
        else:
            ea_vc_data = None
        from ..EA import ea_next_gen
        ea_next_gen.next_gen(self.rin, self.init_struc_data, None,
                             self.opt_struc_data, self.rslt_data, ea_id_data, ea_vc_data)

    def save_id_data(self):
        # ---------- save id_data
        pkl_data.save_id_queueing(self.id_queueing)
        pkl_data.save_id_running(self.id_running)
        if self.rin.algo == 'RS':
            pass
        if self.rin.algo == 'BO':
            pkl_data.save_n_selection(self.n_selection)
            pkl_data.save_id_select_hist(self.id_select_hist)
        if self.rin.algo in ['EA', 'EA-vc']:
            pkl_data.save_gen(self.gen)


    def save_data(self):
        # ---------- save ??_data
        if self.rin.algo == 'BO':
            pkl_data.save_init_dscrpt_data(self.init_dscrpt_data)
            pkl_data.save_opt_dscrpt_data(self.opt_dscrpt_data)
            pkl_data.save_bo_mean(self.bo_mean)
            pkl_data.save_bo_var(self.bo_var)
            pkl_data.save_bo_score(self.bo_score)
        # ea_data is used only in ea_next_gen.py



'''
Initialize CrySPY
'''

from datetime import datetime
from logging import getLogger
import os

import pandas as pd
import pkg_resources

from ..IO import pkl_data, io_stat, write_input
from ..IO.read_input import ReadInput
from ..RS.rs_gen import gen_random
from ..util.utility import get_version
from ..util.struc_util import out_poscar

# ---------- import later
#from ..RS import rs_init
#from ..BO import bo_init
#from ..LAQA import laqa_init
#from ..EA import ea_init


logger = getLogger('cryspy')


def initialize(comm, mpi_rank, mpi_size):
    # ---------- start
    if mpi_rank == 0:
        logger.info('\n\n\nStart CrySPY ' + get_version() + '\n\n')
    # ---------- check versions
    if mpi_rank == 0:
        logger.info('# ---------- Library version info')
        logger.info(f'pandas version: {pd.__version__}')
        v_pymat = pkg_resources.get_distribution('pymatgen').version
        v_pyxtal = pkg_resources.get_distribution('pyxtal').version
        logger.info(f'pymatgen version: {v_pymat}')
        logger.info(f'pyxtal version: {v_pyxtal}')
    # ---------- read input
    if mpi_rank == 0:
        logger.info('# ---------- Read input file, cryspy.in')
    # ########## MPI start
    if mpi_size > 1:
        comm.barrier()
    try:
        rin = ReadInput()    # read input data, cryspy,in
    except Exception as e:
        if mpi_rank == 0:
            logger.error(e)
        raise SystemExit(1)
    # ########## MPI end
    if mpi_rank == 0:
        # ---------- make data directory
        os.makedirs('data/pkl_data', exist_ok=True)
        # ---------- write and save input file
        write_input.out_input(rin)    # log
        pkl_data.save_input(rin)      # input_data.pkl

    # ---------- generate initial structures
    if not rin.load_struc_flag:
        if mpi_size > 1:
            comm.barrier()
        if mpi_rank == 0:
            # ------ time
            time_start = datetime.today()
        # ########## MPI start
        # ------ structure generation
        # only init_struc_data in rank0 is important
        init_struc_data, struc_mol_id = gen_random(rin, rin.tot_struc, 0,
                                                       comm, mpi_rank, mpi_size)
        # ########## MPI end
        if mpi_rank == 0:
            # ------ init_POSCARS
            out_poscar(init_struc_data, './data/init_POSCARS', mode='w')
            # ------ save
            pkl_data.save_init_struc(init_struc_data)
            if rin.algo in ['EA', 'EA-vc'] and rin.struc_mode in ['mol', 'mol_bs']:
                pkl_data.save_struc_mol_id(struc_mol_id)
            # ------ time
            time_end = datetime.today()
            etime = time_end - time_start
            logger.info(f'Elapsed time for structure generation: {etime}')
    else:
        # ------ load initial structure
        if mpi_rank == 0:
            logger.info('# --------- Load initial structure data')
            logger.info('Load ./data/pkl_data/init_struc_data.pkl')
            init_struc_data = pkl_data.load_init_struc()
            # -- check
            if not rin.tot_struc == len(init_struc_data):
                logger.error(f'rin.tot_struc = {rin.tot_struc},'
                                f' len(init_struc_data) = {len(init_struc_data)}')
                raise SystemExit(1)
            # -- init_POSCARS
            out_poscar(init_struc_data, './data/init_POSCARS', mode='w')

    if mpi_rank == 0:
        # ---------- initialize stat
        io_stat.stat_init()

        # ---------- initialize opt_struc_data
        opt_struc_data = {}
        pkl_data.save_opt_struc(opt_struc_data)

        # ---------- initialize rslt_data
        rslt_data = pd.DataFrame(columns=['Spg_num', 'Spg_sym',
                                        'Spg_num_opt', 'Spg_sym_opt',
                                        'E_eV_atom', 'Magmom', 'Opt'])
        rslt_data[['Spg_num', 'Spg_num_opt']] = rslt_data[
                                    ['Spg_num', 'Spg_num_opt']].astype(int)
        pkl_data.save_rslt(rslt_data)

        # ---------- initialize for each algorithm
        if rin.algo == 'RS':
            from ..RS import rs_init
            rs_init.initialize(rin)
        elif rin.algo == 'BO':
            from ..BO import bo_init
            bo_init.initialize(rin, init_struc_data, rslt_data)
        elif rin.algo == 'LAQA':
            from ..LAQA import laqa_init
            laqa_init.initialize(rin)
        elif rin.algo in ['EA', 'EA-vc']:
            from ..EA import ea_init
            ea_init.initialize(rin, init_struc_data, rslt_data)

        # ---------- initialize etc
        if rin.kpt_flag:
            kpt_data = {}
            pkl_data.save_kpt(kpt_data)
        if rin.energy_step_flag:
            energy_step_data = {}
            pkl_data.save_energy_step(energy_step_data)
        if rin.struc_step_flag:
            struc_step_data = {}
            pkl_data.save_struc_step(struc_step_data)
        if rin.force_step_flag:
            force_step_data = {}
            pkl_data.save_force_step(force_step_data)
        if rin.stress_step_flag:
            stress_step_data = {}
            pkl_data.save_stress_step(stress_step_data)

        # ---------- for ext
        if rin.calc_code == 'ext':
            os.makedirs('ext', exist_ok=True)
            with open('ext/stat_job', 'w') as fstat:
                fstat.write('out\n')


from logging import getLogger

import numpy as np
from pymatgen.core import Structure

from . import structure as qe_structure
from ...util import constants
from ...IO import pkl_data


logger = getLogger('cryspy')

def collect_qe(rin, cid, work_path, nat):
    # ---------- check optimization in previous stage
    try:
        with open(work_path+rin.qe_outfile, 'r') as fpout:
            lines = fpout.readlines()
        check_opt = 'not_yet'
        for line in lines:
            if 'End final coordinates' in line:
                check_opt = 'done'
    except Exception as e:
        logger.warning(e)
        check_opt = 'no_file'

    # ---------- obtain energy and magmom
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    try:
        with open(work_path+rin.qe_outfile, 'r') as fpout:
            lines = fpout.readlines()
        energy = np.nan
        for line in reversed(lines):
            if rin.pv_term:
                condition = 'Final enthalpy' in line
            else:
                condition = line.startswith('!')
            if condition:
                energy = float(line.split()[-2])    # in Ry
                energy = energy * constants.RY2EV / float(natot)    # Ry/cell --> eV/atom
                break
        magmom = np.nan    # implemented by H. Sawahata 2020/10/04
        for line in reversed(lines):
            if line.find("total magnetization") >= 0:
                muB = line.split()
                magmom = float(muB[3])
                break
    except Exception as e:
        energy = np.nan    # error
        magmom = np.nan    # error
        logger.warning(f'{e}:    Structure ID {cid},'
                       f'could not obtain energy from {rin.qe_outfile}')

    # ---------- collect the last structure
    try:
        lines_cell = qe_structure.extract_cell_parameters(
            work_path+rin.qe_outfile)
        if lines_cell is None:
            lines_cell = qe_structure.extract_cell_parameters(
                work_path+rin.qe_infile)
        lines_atom = qe_structure.extract_atomic_positions(
            work_path+rin.qe_outfile, nat)
        if lines_atom is None:
            lines_atom = qe_structure.extract_atomic_positions(
                work_path+rin.qe_infile, nat)
        opt_struc = qe_structure.from_lines(lines_cell, lines_atom)

        # ------ opt_qe-structure
        with open('./data/opt_qe-structure', 'a') as fstruc:
            fstruc.write(f'# ID {cid}\n')
        qe_structure.write(opt_struc, './data/opt_qe-structure', mode='a')
    except Exception as e:
        logger.warning(e)
        opt_struc = None

    # ---------- check
    if np.isnan(energy):
        opt_struc = None
    if opt_struc is None:
        energy = np.nan
        magmom = np.nan

    # ---------- return
    return opt_struc, energy, magmom, check_opt


def get_energy_step_qe(rin, energy_step_data, cid, work_path, nat):
    '''
    get energy step data in eV/atom

    energy_step_data[ID][stage][step]
    energy_step_data[ID][0] <-- stage 1
    energy_step_data[ID][1] <-- stage 2
    '''
    # ---------- natot
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    try:
        # ---------- read output file
        with open(work_path+rin.qe_outfile, 'r') as f:
            lines = f.readlines()
        # ---------- get energy step
        energy_step = []
        final_flag = False   # End final coordinates
        vc_flag = False      # vc-relax
        for line in lines:
            if line.startswith('!'):
                energy_step.append(line.split()[4])
            # ------ check opt and vc-relax
            if 'End final coordinates' in line:
                final_flag = True
            if 'CELL_PARAMETERS' in line:
                vc_flag = True
        # ------ delete last energy (after End final coordinates)
        if final_flag and vc_flag:
            energy_step.pop(-1)
        # ------ list --> array, Ry/cell --> eV/atom
        if not energy_step:
            energy_step = None    # if empty
            logger.warning(f'#### ID: {cid}: failed to parse energy_step')
        else:
            energy_step = constants.RY2EV / natot * np.array(energy_step,
                                                               dtype='float')
    except Exception as e:
        energy_step = None
        logger.warning(f'{e}:    ID {cid}: failed to parse energy_step')

    # ---------- append energy_step
    if energy_step_data.get(cid) is None:
        energy_step_data[cid] = []    # initialize
    energy_step_data[cid].append(energy_step)

    # ---------- save energy_step_data
    pkl_data.save_energy_step(energy_step_data)

    # ---------- return
    return energy_step_data


def get_struc_step_qe(rin, struc_step_data, cid, work_path, nat):
    '''
    get structure step data

    # ---------- args
    struc_step_data: (dict) the key is structure ID

    struc_step_data[ID][stage][step]
    struc_step_data[ID][0] <-- stage 1
    struc_step_data[ID][1] <-- stage 2
    '''
    try:
        struc_step = []
        # ------ init struc from pwscf.in
        _extract_struc_qe(work_path+rin.qe_infile, struc_step, nat)
        # ------ struc step from pwscf.out
        _extract_struc_qe(work_path+rin.qe_outfile, struc_step, nat)
        # ------ delete last structure due to duplication
        struc_step.pop(-1)
    except Exception as e:
        struc_step = None
        logger.warning(f'{e}:    ID {cid}: failed to parse in struc_step')

    # ---------- append struc_step_data
    if struc_step_data.get(cid) is None:
        struc_step_data[cid] = []    # initialize
    struc_step_data[cid].append(struc_step)

    # ---------- save struc_step_data
    pkl_data.save_struc_step(struc_step_data)

    # ---------- return
    return struc_step_data


def _extract_struc_qe(filename, struc_step, nat):
    # ---------- read a file
    with open(filename, 'r') as f:
        lines = f.readlines()
    # ---------- extract struc
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    read_cell = False
    read_coords = False
    vc_flag = False      # in case of vc-relax
    for line in lines:
        # ------ cell part
        if read_cell:
            lattice.append(line.split())
            if len(lattice) == 3:
                read_cell = False
                lattice = np.array(lattice, dtype='float')
        if 'CELL_PARAMETERS' in line:
            read_cell = True
            vc_flag = True
            lattice = []
        # ------ coords part
        if read_coords:
            lsplit = line.split()
            species.append(lsplit[0])
            coords.append(lsplit[1:])
            if len(coords) == natot:
                read_coords = False
                coords = np.array(coords, dtype='float')
                # ---- gen struc
                if not vc_flag:    # empty lattice, use init lattice
                    lattice = struc_step[0].lattice
                struc = Structure(lattice, species, coords)
                struc_step.append(struc)
        if 'ATOMIC_POSITIONS' in line:
            read_coords = True
            species = []
            coords = []


def get_force_step_qe(rin, force_step_data, cid, work_path, nat):
    '''
    get force step data in eV/angstrom

    # ---------- args
    force_step_data: (dict) the key is structure ID

    force_step_data[ID][stage][step]
    force_step_data[ID][0] <-- stage 1
    force_step_data[ID][1] <-- stage 2
    '''
    # ---------- natot
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    try:
        # ---------- read output file
        with open(work_path+rin.qe_outfile, 'r') as f:
            lines = f.readlines()
        # ---------- get force step
        force_step = []
        read_force = False
        final_flag = False   # End final coordinates
        vc_flag = False      # in case of vc-relax
        for line in lines:
            if 'atom    1 type' in line:
                read_force = True
                force = []
            if read_force:
                force.append(line.split()[6:])
                if len(force) == natot:
                    read_force = False
                    force_step.append(constants.RY2EV / constants.BOHR2ANG * np.array(
                        force, dtype='float'))
            # ------ check opt and vc-relax
            if 'End final coordinates' in line:
                final_flag = True
            if 'CELL_PARAMETERS' in line:
                vc_flag = True
        # ------ delete last energy (after End final coordinates)
        if final_flag and vc_flag:
            force_step.pop(-1)
        # ------ if empty
        if len(force_step) == 0:
            force_step = None
            logger.warning(f'#### ID: {cid}: failed to parse force_step')
    except Exception as e:
        force_step = None
        logger.warning(e + f'#### ID: {cid}: failed to parse in force_step')

    # ---------- append force_step
    if force_step_data.get(cid) is None:
        force_step_data[cid] = []    # initialize
    force_step_data[cid].append(force_step)

    # ---------- save force_step_data
    pkl_data.save_force_step(force_step_data)

    # ---------- return
    return force_step_data


def get_stress_step_qe(rin, stress_step_data, cid, work_path):
    '''
    get stress step data in eV/ang**3

    # ---------- args
    stress_step_data: (dict) the key is structure ID

    stress_step_data[ID][stage][step]
    stress_step_data[ID][0] <-- stage 1
    stress_step_data[ID][1] <-- stage 2
    '''
    try:
        # ---------- read output file
        with open(work_path+rin.qe_outfile, 'r') as f:
            lines = f.readlines()
        # ---------- get stress step
        stress_step = []
        read_stress = False
        final_flag = False   # End final coordinates
        vc_flag = False      # in case of vc-relax
        for line in lines:
            if read_stress:
                stress.append(line.split()[3:])
                if len(stress) == 3:
                    read_stress = False
                    stress_step.append(constants.KBAR2eV_ANG3 * np.array(
                        stress, dtype='float'))
            if 'total   stress  (Ry/bohr**3)' in line:
                read_stress = True
                stress = []
                # ------ check opt and vc-relax
            if 'End final coordinates' in line:
                final_flag = True
            if 'CELL_PARAMETERS' in line:
                vc_flag = True
        # ------ delete last energy (after End final coordinates)
        if final_flag and vc_flag:
            stress_step.pop(-1)
        # ------ if empty
        if len(stress_step) == 0:
            stress_step = None
            logger.warning(f'#### ID: {cid}: failed to parse stress_step')
    except Exception as e:
        stress_step = None
        logger.warning(e + f'#### ID: {cid}: failed to parse in stress_step')

    # ---------- append stress_step
    if stress_step_data.get(cid) is None:
        stress_step_data[cid] = []    # initialize
    stress_step_data[cid].append(stress_step)

    # ---------- save stress_step_data
    pkl_data.save_stress_step(stress_step_data)

    # ---------- return
    return stress_step_data



# from ase.build import bulk
import os
import pickle
from pymatgen.io.ase import AseAtomsAdaptor


def get_ase_atoms(status, cid):

    path = f'./data/pkl_data/{status}_struc_data.pkl'

    if os.path.isfile(path):
        with open(path, 'rb') as f:
            struc_data = pickle.load(f)
        struc = struc_data[cid]
        atoms = AseAtomsAdaptor.get_atoms(struc)
        return atoms
    else:
        print("This structure does not exist.")



from logging import getLogger
import os


logger = getLogger('cryspy')


def check_input_lammps(rin):
    # ---------- prepare rin.jobfile, rin.lammps_potential, rin.rammps_infile
    if rin.lammps_potential is None:
        calc_inputs = [rin.jobfile, rin.lammps_infile]
    else:
        calc_inputs = [rin.jobfile, rin.lammps_infile] + rin.lammps_potential

    # ----- check required files
    for f in calc_inputs:
        if f == rin.lammps_infile:
            finfiles = [rin.lammps_infile + f'_{i}' for i in range(
                1, rin.nstage+1)]
            for ff in finfiles:
                if not os.path.isfile('./calc_in/'+ff):
                    logger.error('Could not find ./calc_in/'+ff)
                    raise SystemExit(1)
        else:
            if not os.path.isfile('./calc_in/'+f):
                logger.error('Could not find ./calc_in/'+f)
                raise SystemExit(1)


import configparser
from contextlib import redirect_stdout
from logging import getLogger
import os

import numpy as np
import physbo

from ..IO import io_stat, out_results, pkl_data


logger = getLogger('cryspy')


def next_select(rin, rslt_data, bo_id_data, bo_data, noprint=False):
    # ---------- log
    logger.info('# ------ Bayesian optimization')

    # ---------- bo_id_data and bo_data
    n_selection, id_running, id_queueing, id_select_hist = bo_id_data
    init_dscrpt_data, opt_dscrpt_data, bo_mean, bo_var, bo_score = bo_data

    # ---------- n_selection
    n_selection += 1

    # ---------- manual select
    if rin.manual_select_bo is not None:
        manual_list = list(rin.manual_select_bo)
        x = ' '.join(str(i) for i in manual_list)
        logger.info(f'Manual select: {x}')
        # ------ check already selected
        tmp_list = [i for i in manual_list if i in opt_dscrpt_data]
        for j in tmp_list:
            manual_list.remove(j)
            logger.info(f'ID {j} was already selected. Remove it')
        # ------ number of structure to be selected
        nselect = rin.nselect_bo - len(manual_list)
        id_queueing = manual_list
        # ------ delete the value for manual_select_bo in cryspy.in
        config = configparser.ConfigParser()
        config.read('cryspy.in')
        config.set('BO', 'manual_select_bo', '')    # clear
        with open('cryspy.in', 'w') as f:
            config.write(f)
    else:
        nselect = rin.nselect_bo
    # ------ if number of remaining structures < nselect
    if len(init_dscrpt_data) - len(opt_dscrpt_data) < nselect:
        nselect = len(init_dscrpt_data) - len(opt_dscrpt_data)

    # ---------- selection
    if 0 < nselect:
        # ------ descriptors
        descriptors = []
        s_act = []
        done_id = []    # finished and non_error
        non_error_id = [i for i in range(len(init_dscrpt_data))]
        for i in range(len(init_dscrpt_data)):
            if i in opt_dscrpt_data:
                # -- already done
                if opt_dscrpt_data[i] is None:    # find error
                    non_error_id.remove(i)
                    continue
                x = rslt_data.loc[i]['E_eV_atom']
                if rin.emax_bo is not None:
                    if x > rin.emax_bo:
                        non_error_id.remove(i)
                        logger.info(f'Eliminate ID {i}: {x} > emax_bo')
                        continue
                if rin.emin_bo is not None:
                    if x < rin.emin_bo:
                        non_error_id.remove(i)
                        logger.info(f'Eliminate ID {i}: {x} < emin_bo')
                        continue
                s_act.append(len(descriptors))
                done_id.append(i)
                descriptors.append(opt_dscrpt_data[i])
            else:
                # -- not yet
                if rin.manual_select_bo is not None:
                    if i in rin.manual_select_bo:
                        # remove manualy selected IDs not to select them
                        non_error_id.remove(i)
                    else:
                        descriptors.append(init_dscrpt_data[i])
                else:
                    descriptors.append(init_dscrpt_data[i])
        # ------ targets
        targets = [rslt_data.loc[cid]['E_eV_atom'] for cid in done_id]
        # ------ list --> numpy
        s_act = np.array(s_act, dtype=int)
        descriptors = np.array(descriptors)
        targets = np.array(targets, dtype=float)
        # ------ Bayesian optimization
        actions, cryspy_mean, cryspy_var, cryspy_score = _bayes_opt(
            s_act, descriptors, targets, nselect,
            rin.score, rin.cdev, rin.num_rand_basis, noprint)
        # t_act, t_mean, t_var, t_score = _bayes_opt2(
        #     s_act, descriptors, targets, nselect,
        #     rin.score, rin.cdev, rin.num_rand_basis, noprint)
        # logger.debug(f't_act: {t_act}')
        # logger.debug(f't_mean: {t_mean}')
        # logger.debug(f't_var: {t_var}')
        # logger.debug(f't_score: {t_score}')
        # ------ actions --> id_queueing
        for i in actions:
            id_queueing.append(non_error_id[i])
        # ------ bo_mean, bo_var, bo_score
        remaining_id = list(set(non_error_id) - set(done_id))
        bo_mean[n_selection] = dict(zip(remaining_id, cryspy_mean))
        bo_var[n_selection] = dict(zip(remaining_id, cryspy_var))
        if rin.score == 'TS':
            # -- in TS, nested list [[score1, socre2, ...]] is used
            bo_score[n_selection] = dict(zip(remaining_id, cryspy_score[0]))
        else:
            bo_score[n_selection] = dict(zip(remaining_id, cryspy_score))
        # ------ out
        out_results.out_bo_common('bo_mean', bo_mean, rin.tot_struc)
        out_results.out_bo_common('bo_var', bo_var, rin.tot_struc)
        out_results.out_bo_common('bo_score', bo_score, rin.tot_struc)
        out_results.out_bo_status(bo_mean, bo_var, bo_score,
                                  n_selection)
        # ------ save bo_data
        pkl_data.save_init_dscrpt_data(init_dscrpt_data)
        pkl_data.save_opt_dscrpt_data(opt_dscrpt_data)
        pkl_data.save_bo_mean(bo_mean)
        pkl_data.save_bo_var(bo_var)
        pkl_data.save_bo_score(bo_score)

    # ---------- id_select_hist
    id_select_hist.append(id_queueing[:])    # append shallow copy
    out_results.out_bo_id_hist(id_select_hist)

    # ---------- save
    pkl_data.save_id_queueing(id_queueing)
    # pkl_data.save_id_running(id_running)    # not used here
    pkl_data.save_n_selection(n_selection)
    pkl_data.save_id_select_hist(id_select_hist)

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_common(stat, 'selection', n_selection)
    io_stat.set_id(stat, 'selected_id', id_queueing)
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- out and log
    logger.info(f'# ---------- Selection: {n_selection}')
    x = ' '.join(str(a) for a in id_queueing)
    logger.info(f'selected_id: {x}')

    # ---------- ext
    if rin.calc_code == 'ext':
        with open('ext/stat_job', 'w') as fstat:
            fstat.write('out\n')


def _bayes_opt(
    s_act,
    descriptors,
    targets,
    nselect,
    score,
    cdev=0.001,
    num_rand_basis=0,
    noprint=False,
):
    '''
    # ---------- args
    s_act (ndarray): indexes of already calculated structures. NOT structure ID.
    descriptors (ndarray): descriptors of all structures, without error
    targets (ndarray): energies of already calculated structures
    nselect (int): number of structures to be selected
    score (str): 'PI', EI' or 'TS'
    cdev (float): cutoff value for standard deviation
    num_rand_basis (int): number of basis functions. if 0, gaussian process
    noprint (bool): if True, suppress print

    # ---------- return
    actions (ndarray): indexes of selected structures
    p_mean (ndarray): mean of predicted energies
    p_var (ndarray): variance of predicted energies
    p_score (ndarray): value of acquisition function
    '''

    # ---------- standardization
    dev = np.std(descriptors, axis=0)
    nonzero_indx = np.where(dev > cdev)[0]
    X = (descriptors[:, nonzero_indx] - np.mean(
        descriptors[:, nonzero_indx], axis=0)) / dev[nonzero_indx]

    # ---------- Declaring the policy by
    policy = physbo.search.discrete.policy(test_X=X, initial_data=[s_act, -targets])

    # ---------- Bayes search
    if noprint:
        with redirect_stdout(open(os.devnull, 'w')):
            actions = policy.bayes_search(
                                max_num_probes=1,
                                num_search_each_probe=nselect,
                                simulator=None,
                                score=score,
                                interval=0,
                                num_rand_basis = num_rand_basis,
                            )
    else:
        actions = policy.bayes_search(
                            max_num_probes=1,
                            num_search_each_probe=nselect,
                            simulator=None,
                            score=score,
                            interval=0,
                            num_rand_basis = num_rand_basis,
                        )

    # ---------- mean, var, score
    p_mean = policy.get_post_fmean(X)
    p_var = policy.get_post_fcov(X)
    p_score = policy.get_score(mode=score, xs=X)

    # ------ return
    return actions, p_mean, p_var, p_score


'''
Collect results in soiap
'''

from logging import getLogger

import numpy as np

from . import structure as soiap_structure
from ...util import constants
from ...IO import pkl_data


logger = getLogger('cryspy')

def collect_soiap(rin, cid, work_path, nat):
    # ---------- check optimization in current stage
    try:
        with open(work_path+rin.soiap_outfile, 'r') as fout:
            lines = fout.readlines()
        check_opt = 'not_yet'
        for i, line in enumerate(lines):
            if '*** QMD%loopc' in line:
                if ('QMD%frc converged.' in lines[i-2]
                        and 'QMD%strs converged.' in lines[i-1]):
                    check_opt = 'done'
                break
    except Exception:
        check_opt = 'no_file'

    # ---------- obtain energy and magmom
    magmom = np.nan    # magnetic moment is not calculated
    try:
        with open(work_path+'log.tote') as f:
            lines = f.readlines()
        energy = float(lines[-1].split()[4])    # in Hartree/atom
        energy = energy * constants.HRT2EV        # Hartree/atom to eV/atom
    except Exception as e:
        energy = np.nan    # error
        logger.warning(f'{e}:    Structure ID {cid},'
                    f' could not obtain energy from {rin.soiap_outfile}')

    # ---------- collect the last structure
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    try:
        with open(work_path+'log.struc', 'r') as f:
            lines = f.readlines()
            lines = lines[-(natot+5):]
        opt_struc = soiap_structure.from_file(rin, lines, nat)
    except Exception:
        opt_struc = None

    # ---------- check
    if np.isnan(energy):
        opt_struc = None
    if opt_struc is None:
        energy = np.nan
        magmom = np.nan

    # ---------- return
    return opt_struc, energy, magmom, check_opt


def get_energy_step_soiap(energy_step_data, cid, work_path):
    '''
    get energy step data in eV/atom

    energy_step_data[ID][stage][step]
    energy_step_data[ID][0] <-- stage 1
    energy_step_data[ID][1] <-- stage 2

    In soiap, collect energy step data only when loopa == 1.
        This is because other data (struc, force, stress)
        are output only when loopa == 1
        see, https://github.com/nbsato/soiap/blob/master/doc/instructions.md
    '''
    try:
        energy_step = []
        with open(work_path+'log.tote') as f:
            lines = f.readlines()
        for line in lines:
            if line.split()[1] == '1':
                energy_step.append(line.split()[4])    # collumn 4: Hartree/atom
        energy_step = np.array(energy_step, dtype='float') * constants.HRT2EV
    except Exception as e:
        energy_step = None
        logger.warning(f'{e}:    ID {cid}: failed to parse log.tote')

    # ---------- append energy_step
    if energy_step_data.get(cid) is None:
        energy_step_data[cid] = []    # initialize
    energy_step_data[cid].append(energy_step)

    # ---------- save energy_step_data
    pkl_data.save_energy_step(energy_step_data)

    # ---------- return
    return energy_step_data


def get_struc_step_soiap(rin, struc_step_data, cid, work_path, nat):
    '''
    get structure step data

    # ---------- args
    struc_step_data: (dict) the key is structure ID

    struc_step_data[ID][stage][step]
    struc_step_data[ID][0] <-- stage 1
    struc_step_data[ID][1] <-- stage 2
    '''
    # ---------- get struc step from log.struc
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    try:
        # ------ read file
        with open(work_path+'log.struc', 'r') as f:
            lines = f.readlines()
        # ------ init.
        struc_step = []
        # ------ loop for relaxation step
        tmp_lines = []
        for line in lines:
            tmp_lines.append(line)
            if len(tmp_lines) == natot + 5:
                struc = soiap_structure.from_file(rin, tmp_lines, nat)
                struc_step.append(struc)
                tmp_lines = []    # clear
    except Exception as e:
        struc_step = None
        logger.warning(f'{e}    ID {cid}: failed to parse log.struc')

    # ---------- append struc_step
    if struc_step_data.get(cid) is None:
        struc_step_data[cid] = []    # initialize
    struc_step_data[cid].append(struc_step)

    # ---------- save struc_step_data
    pkl_data.save_struc_step(struc_step_data)

    # ---------- return
    return struc_step_data


def get_force_step_soiap(force_step_data, cid, work_path, nat):
    '''
    get force step data in eV/angstrom

    # ---------- args
    force_step_data: (dict) the key is structure ID

    force_step_data[ID][stage][step]
    force_step_data[ID][0] <-- stage 1
    force_step_data[ID][1] <-- stage 2
    '''
    # ---------- get force step from log.frc
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    try:
        # ------ read file
        with open(work_path+'log.frc', 'r') as f:
            lines = f.readlines()
        # ------ init
        force_step = []
        tmp_lines = []
        # ------ parse
        for line in lines:
            if 'forces' not in line:
                tmp_lines.append([float(x) for x in line.split()])
                if len(tmp_lines) == natot:
                    tmp_lines = np.array(tmp_lines)
                    # Hartree/Bohr --> eV/ang
                    tmp_lines = tmp_lines * constants.HRT2EV / constants.BOHR2ANG
                    force_step.append(tmp_lines)
                    tmp_lines = []    # clear
    except Exception as e:
        force_step = None
        logger.warning(f'{e}    ID {cid}: failed to parse log.frc')

    # ---------- append force_step
    if force_step_data.get(cid) is None:
        force_step_data[cid] = []    # initialize
    force_step_data[cid].append(force_step)

    # ---------- save force_step_data
    pkl_data.save_force_step(force_step_data)

    # ---------- return
    return force_step_data


def get_stress_step_soiap(stress_step_data, cid, work_path):
    '''
    get stress step data in eV/ang**3

    # ---------- args
    stress_step_data: (dict) the key is structure ID

    stress_step_data[ID][stage][step]
    stress_step_data[ID][0] <-- stage 1
    stress_step_data[ID][1] <-- stage 2
    '''
    # ---------- get stress step from log.strs
    try:
        # ------ read file
        with open(work_path+'log.strs', 'r') as f:
            lines = f.readlines()
        # ------ init
        stress_step = []
        tmp_lines = []
        # ------ parse
        for line in lines:
            if 'QMD' not in line:
                tmp_lines.append([float(x) for x in line.split()])
                if len(tmp_lines) == 3:
                    tmp_lines = np.array(tmp_lines)
                    # Hartree/Bohr**3 --> eV/ang**3
                    tmp_lines = tmp_lines * constants.HRT2EV / constants.BOHR2ANG**3
                    stress_step.append(tmp_lines)
                    tmp_lines = []    # clear
    except Exception as e:
        stress_step = None
        logger.warning(f'{e}:    ID {cid}: failed to parse log.strs')

    # ---------- append stress_step
    if stress_step_data.get(cid) is None:
        stress_step_data[cid] = []    # initialize
    stress_step_data[cid].append(stress_step)

    # ---------- save stress_step_data
    pkl_data.save_stress_step(stress_step_data)

    # ---------- return
    return stress_step_data



import configparser
from dataclasses import dataclass, field
from logging import getLogger
import os

from ..util import utility


logger = getLogger('cryspy')


@dataclass
class ReadInput:
    '''
    Read cryspy.in
        Here, ignore type hint for None
    '''
    # ---------- basic section
    algo: str = field(default=None)
    calc_code: str = field(default=None)
    tot_struc: int = field(default=None)
    nstage: int = field(default=None)
    njob: int = field(default=None)
    jobcmd: str = field(default=None)
    jobfile: str = field(default=None)

    # ---------- structure section
    struc_mode: str = field(default=None)
    natot: int = field(default=None)
    atype: tuple = field(default=None)
    nat: tuple = field(default=None)
    mindist: tuple = field(default=None)
    mindist_factor: float = field(default=None)
    vol_factor: float = field(default=None)
    vol_mu: float = field(default=None)
    vol_sigma: float = field(default=None)
    symprec: float = field(default=None)
    spgnum: str = field(default=None)    # ignore type hint here. 'all', tuple or 0
    use_find_wy: bool = field(default=None)
    # ------ EA-vc
    ll_nat: tuple = field(default=None)
    ul_nat: tuple = field(default=None)
    # ------ mol or mol_bs
    mol_file: tuple = field(default=None)
    nmol: tuple = field(default=None)
    timeout_mol: float = field(default=None)
    rot_mol: str = field(default=None)
    nrot: int = field(default=None)
    mindist_mol_bs: tuple = field(default=None)
    mindist_mol_bs_factor: float = field(default=None)
    # ------ find_wy or spgnum = 0
    fwpath: str = field(default=None)
    minlen: float = field(default=None)
    maxlen: float = field(default=None)
    dangle: float = field(default=None)
    maxcnt: int = field(default=None)

    # ---------- option section
    stop_chkpt: int = field(default=None)
    load_struc_flag: bool = field(default=None)
    stop_next_struc: bool = field(default=None)
    recalc: tuple = field(default=None)
    append_struc_ea: bool = field(default=None)
    energy_step_flag: bool = field(default=None)
    struc_step_flag: bool = field(default=None)
    force_step_flag: bool = field(default=None)
    stress_step_flag: bool = field(default=None)

    # ---------- BO section
    nselect_bo: int = field(default=None)
    score: str = field(default=None)
    num_rand_basis: int = field(default=None)
    cdev: float = field(default=None)
    dscrpt: str = field(default=None)
    fp_rmin: float = field(default=None)
    fp_rmax: float = field(default=None)
    fp_npoints: int = field(default=None)
    fp_sigma: float = field(default=None)
    max_select_bo: int = field(default=None)
    manual_select_bo: tuple = field(default=None)
    emin_bo: float = field(default=None)
    emax_bo: float = field(default=None)

    # ---------- LAQA section
    nselect_laqa: int = field(default=None)
    wf: float = field(default=None)
    ws: float = field(default=None)

    # ---------- EA section
    n_pop: int = field(default=None)
    n_crsov: int = field(default=None)
    n_perm: int = field(default=None)
    n_strain: int = field(default=None)
    n_rand: int = field(default=None)
    n_elite: int = field(default=None)
    fit_reverse: bool = field(default=None)
    n_fittest: int = field(default=None)
    slct_func: str = field(default=None)
    t_size: int = field(default=None)
    a_rlt: float = field(default=None)
    b_rlt: float = field(default=None)
    crs_lat: str = field(default=None)
    nat_diff_tole: int = field(default=None)
    ntimes: int = field(default=None)
    sigma_st: float = field(default=None)
    maxcnt_ea: int = field(default=None)
    maxgen_ea: int = field(default=None)
    emin_ea: float = field(default=None)
    emax_ea: float = field(default=None)
    n_add: int = field(default=None)
    n_elim: int = field(default=None)
    n_subs: int = field(default=None)
    target: str = field(default=None)
    end_point: tuple = field(default=None)
    n_rotation: int = field(default=None)
    mindist_mol_ea: tuple = field(default=None)    # not implemented yet, for EA mol
    rot_max_angle: float = field(default=None)     # not implemented yet, for EA mol
    protect_mol_struc: bool = field(default=None)  # not implemented yet, for EA mol

    # ---------- common in VASP, QE, OMX
    kppvol: int = field(default=None)
    kpt_flag: bool = field(default=None)
    force_gamma: bool = field(default=None)

    # ---------- VASP section
    # no inputs other than common

    # ---------- QE section
    qe_infile: str = field(default=None)
    qe_outfile: str = field(default=None)
    pv_term: bool = field(default=None)

    # ---------- OMX section
    OMX_infile: str = field(default=None)
    OMX_outfile: str = field(default=None)
    upSpin: dict = field(default=None)
    downSpin: dict = field(default=None)

    # ---------- soiap section
    soiap_infile: str = field(default=None)
    soiap_outfile: str = field(default=None)
    soiap_cif: str = field(default=None)

    # ---------- lammps section
    lammps_infile: str = field(default=None)
    lammps_outfile: str = field(default=None)
    lammps_potential: str = field(default=None)
    lammps_data: str = field(default=None)

    # ---------- ASE section
    ase_python: str = field(default=None)

    # ---------- ext
    # no inputs other than kpt_flag

    def __post_init__(self):
        self._read_cryspyin()    # self.config is created
        self._read_basic()       # read basic section in cryspy.in
        self._read_structure()   # read structure section in cryspy.in
        self._read_option()      # read option section in cryspy.in, do this prior to EA for append_struc_ea
        if self.algo == 'BO':
            self._read_bo()      # read BO section in cryspy.in
        if self.algo == 'LAQA':
            self._read_laqa()    # read LAQA section in cryspy.in
        if self.algo in ['EA', 'EA-vc'] or self.append_struc_ea:
            self._read_ea()      # read EA section in cryspy.in
        if self.calc_code == 'VASP':
            self._read_vasp()    # read VASP section in cryspy.in
        if self.calc_code == 'QE':
            self._read_qe()      # read QE section in cryspy.in
        if self.calc_code == 'OMX':
            self._read_omx()     # read OMX section in cryspy.in
        if self.calc_code == 'soiap':
            self._read_soiap()   # read soiap section in cryspy.in
        if self.calc_code == 'LAMMPS':
            self._read_lammps()  # read LAMMPS section in cryspy.in
        if self.calc_code == 'ASE':
            self._read_ase()     # read ASE section in cryspy.in
        if self.calc_code == 'ext':
            self._read_ext()     # read ext section in cryspy.in

    def _read_cryspyin(self):
        filename = 'cryspy.in'
        if not os.path.isfile(filename):
            raise FileNotFoundError(f'{filename} not found in {os.getcwd()}')
        self.config = configparser.ConfigParser()
        self.config.read(filename)

    def _read_basic(self):
        # ---------- algo
        self.algo = self.config.get('basic', 'algo')
        if self.algo not in ['RS', 'BO', 'LAQA', 'EA', 'EA-vc']:
            raise ValueError('algo must be RS, BO, LAQA, EA or EA-vc')
        # ---------- calc_code
        self.calc_code = self.config.get('basic', 'calc_code')
        if self.calc_code not in ['VASP', 'QE', 'soiap', 'LAMMPS', 'OMX', 'ASE', 'ext']:
            raise ValueError('calc_code must be VASP, QE, OMX, soiap, LAMMPS, ASE, or ext')
        if self.algo == 'LAQA':
            if self.calc_code not in ['VASP', 'QE', 'soiap']:
                raise ValueError('LAQA: only VASP, QE, or soiap for now')
        # ---------- tot_struc
        self.tot_struc = self.config.getint('basic', 'tot_struc')
        if self.tot_struc < 1:
            raise ValueError('tot_struc < 1, check tot_struc')
        # ---------- nstage
        if not self.calc_code == 'ext':
            self.nstage = self.config.getint('basic', 'nstage')
            if self.nstage < 1:
                raise ValueError('nstage < 1, check nstage')
            if self.algo == 'LAQA':
                if not self.nstage == 1:
                    raise ValueError('nstage must be 1 in LAQA')
        # ---------- njob
        if not self.calc_code == 'ext':
            self.njob = self.config.getint('basic', 'njob')
            if self.njob < 1:
                raise ValueError('njob < 1, check njob')
        # ---------- jobcmd
        if not self.calc_code == 'ext':
            self.jobcmd = self.config.get('basic', 'jobcmd')
        # ---------- jobfile
        if not self.calc_code == 'ext':
            self.jobfile = self.config.get('basic', 'jobfile')

    def _read_structure(self):
        # ---------- struc_mode
        try:
            self.struc_mode = self.config.get('structure', 'struc_mode')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.struc_mode = 'crystal'
        if self.struc_mode not in ['crystal', 'mol', 'mol_bs']:
            raise ValueError('struc_mode must be crystal, mol, or  mol_bs')
        #if self.struc_mode not in ['crystal', 'mol', 'mol_bs', 'host']:
        #    raise ValueError('struc_mode must be crystal, mol, mol_bs, or host')
        if self.algo in ['EA', 'EA-vc'] and self.struc_mode in ['mol', 'mol_bs']:
            if self.calc_code == 'ext':
                raise ValueError('EA, mol or mol_bs, ext, not yet')
        # ---------- natot
        if not self.algo == 'EA-vc':
            self.natot = self.config.getint('structure', 'natot')
            if self.natot < 1:
                raise ValueError('natot < 1, check natot')
        # ---------- atype
        self.atype = self.config.get('structure', 'atype')
        self.atype = tuple([a for a in self.atype.split()])    # str --> list --> tuple
        if self.algo =='EA-vc':
            if not len(self.atype) == 2:
                raise ValueError('len(atype) must be 2 in EA-vc for now')
        # ---------- nat
        if not self.algo == 'EA-vc':
            self.nat = self.config.get('structure', 'nat')
            self.nat = tuple([int(x) for x in self.nat.split()])    # str --> int --> list --> tuple
            if not len(self.nat) == len(self.atype):
                raise ValueError('not len(nat) == len(atype), check atype and nat')
            if not sum(self.nat) == self.natot:
                raise ValueError('not sum(nat) == natot, check natot and nat')
        # ---------- mindist
        try:
            self.mindist = []
            for i in range(len(self.atype)):
                tmp = self.config.get('structure', f'mindist_{i+1}')
                tmp = tuple([float(x) for x in tmp.split()])    # str --> float --> list --> tuple
                if not len(tmp) == len(self.atype):
                    raise ValueError(f'not len(mindist_{i+1}) == len(atype)')
                self.mindist.append(tmp)
            self.mindist = tuple(self.mindist)    # list --> tuple
            # ------ check symmetric matrix
            for i in range(len(self.mindist)):
                for j in range(len(self.mindist)):
                    if i < j:
                        if not self.mindist[i][j] == self.mindist[j][i]:
                            raise ValueError(f'mindist is not symmetric. ({i}, {j}) -->'
                                            f' {self.mindist[i][j]}, ({j}, {i}) --> {self.mindist[j][i]}')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.mindist = None
        # ---------- mindist_factor
        try:
            self.mindist_factor = self.config.getfloat('structure', 'mindist_factor')
            if self.mindist_factor <= 0.0:
                raise ValueError('mindist_factor must be positive')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.mindist_factor = 1.0
        # ---------- vol_factor
        try:
            self.vol_factor = self.config.getfloat('structure', 'vol_factor')
            if self.vol_factor <= 0.0:
                raise ValueError('vol_factor must be positive')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.vol_factor = 1.1
        # ---------- vol_mu
        try:
            self.vol_mu = self.config.getfloat('structure', 'vol_mu')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.vol_mu = None
        if self.vol_mu is not None:
            if self.vol_mu <= 0.0:
                raise ValueError('vol_mu must be positive float')
        # ---------- vol_sigma
        try:
            self.vol_sigma = self.config.getfloat('structure', 'vol_sigma')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.vol_sigma = None
        if self.vol_mu is not None and self.vol_sigma is None:
            raise ValueError('vol_mu is set, but vol_sigma is not set')
        if self.vol_sigma is not None:
            if self.vol_sigma < 0.0:
                raise ValueError('vol_sigma must not be negative')
        # ---------- symprec
        try:
            self.symprec = self.config.getfloat('structure', 'symprec')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.symprec = 0.01
        if self.symprec < 0.0:
            raise ValueError('symprec must be positive float')
        # ---------- spgnum
        try:
            self.spgnum = self.config.get('structure', 'spgnum')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.spgnum = 'all'
        if self.spgnum == '0':    # str here
            if self.struc_mode in ['mol', 'mol_bs']:
                raise ValueError('spgnum = 0 is not allow when struc_mode is mol or mol_bs')
            self.spgnum = 0    # int here
        elif self.spgnum == 'all':
            pass
        else:
            self._spgtuple()    # self.spgnum: e.g. 3 15-18 100 --> (3, 15, 16, 17, 18, 100)
        # ---------- use_find_wy
        try:
            self.use_find_wy = self.config.getboolean('structure', 'use_find_wy')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.use_find_wy = False
        if self.use_find_wy:
            if not self.struc_mode == 'crystal':
                raise ValueError('find_wy can be use if struc_mode is crystal')
        # ---------- EA-vc
        # ------ ll_nat, ul_nat
        # --
        if self.algo =='EA-vc':
            self.ll_nat = self.config.get('structure', 'll_nat')
            self.ll_nat = tuple([int(x) for x in self.ll_nat.split()])    # str --> int --> list --> tuple
            self.ul_nat = self.config.get('structure', 'ul_nat')
            self.ul_nat = tuple([int(x) for x in self.ul_nat.split()])    # str --> int --> list --> tuple
            if not len(self.atype) == len(self.ll_nat) == len(self.ul_nat):
                raise ValueError('not len(atype) == len(ll_nat) == len(ul_nat), check ll_nat and ul_nat')
            for i in range(len(self.ll_nat)):
                if not 1 <= self.ll_nat[i] <= self.ul_nat[i]:
                    raise ValueError(f'not 1 <= ll_nat[{i}] <= ul_nat[{i}], check ll_nat and ul_nat')
        # ---------- mol or mol_bs
        if self.struc_mode in ['mol', 'mol_bs']:
            # ------ mol_file
            self.mol_file = self.config.get('structure', 'mol_file')
            self.mol_file = tuple([a for a in self.mol_file.split()])    # str --> list --> tuple
            # ------ nmol
            self.nmol = self.config.get('structure', 'nmol')
            self.nmol = tuple([int(x) for x in self.nmol.split()])    # str --> int --> list --> tuple
            if not len(self.mol_file) == len(self.nmol):
                raise ValueError('not len(mol_file) == len(nmol)')
            # ------ timeout_mol
            try:
                self.timeout_mol = self.config.getfloat('structure', 'timeout_mol')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.timeout_mol = None
            if self.timeout_mol is not None:
                if self.timeout_mol <= 0:
                    raise ValueError('timeout_mol must be None or positive')
        # ---------- mol_bs
        if self.struc_mode == 'mol_bs':
            # ------ rot_mol
            try:
                self.rot_mol = self.config.get('structure', 'rot_mol')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.rot_mol = 'random_wyckoff'
            if self.rot_mol not in ['random', 'random_mol', 'random_wyckoff']:
                raise ValueError('rot_mol must be random, random_mol, or random_wyckoff')
            # ------ nrot
            try:
                self.nrot = self.config.getint('structure', 'nrot')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.nrot = 20
            if self.nrot < 1:
                raise ValueError('nrot < 1, check nrot')
            # ------ mindist_mol_bs
            try:
                self.mindist_mol_bs = []
                for i in range(len(self.mol_file)):
                    tmp = self.config.get('structure', f'mindist_mol_bs_{i+1}')
                    tmp = tuple([float(x) for x in tmp.split()])    # str --> float --> list --> tuple
                    if not len(tmp) == len(self.mol_file):
                        raise ValueError(f'not len(mindist_mol_bs_{i+1}) == len(mol_file)')
                    self.mindist_mol_bs.append(tmp)
                self.mindist_mol_bs = tuple(self.mindist_mol_bs)    # list --> tuple
                # -- check symmetric matrix
                for i in range(len(self.mindist_mol_bs)):
                    for j in range(len(self.mindist_mol_bs)):
                        if i < j:
                            if not self.mindist_mol_bs[i][j] == self.mindist_mol_bs[j][i]:
                                raise ValueError(f'mindist_mol_bs is not symmetric. ({i}, {j}) -->'
                                                f' {self.mindist_mol_bs[i][j]}, ({j}, {i}) --> {self.mindist_mol_bs[j][i]}')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.mindist_mol_bs = None
            # ------ mindist_mol_bs_factor
            try:
                self.mindist_mol_bs_factor = self.config.getfloat('structure', 'mindist_mol_bs_factor')
                if self.mindist_mol_bs_factor <= 0.0:
                    raise ValueError('mindist_mol_bs_factor must be positive')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.mindist_mol_bs_factor = 1.0
        # ---------- find_wy or spgnum = 0
        if self.spgnum == 0 or self.use_find_wy:
            # ------ fwpath
            try:
                self.fwpath = self.config.get('structure', 'fwpath')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.fwpath = None
            # ------ minlen
            self.minlen = self.config.getfloat('structure', 'minlen')
            if self.minlen <= 0.0:
                raise ValueError('minlen must be positive')
            # ------ maxlen
            self.maxlen = self.config.getfloat('structure', 'maxlen')
            if self.minlen > self.maxlen:
                raise ValueError('minlen > maxlen')
            # ------ dangle
            self.dangle = self.config.getfloat('structure', 'dangle')
            if self.dangle <= 0.0:
                raise ValueError('dangle must be positive')
            # ------ maxcnt
            try:
                self.maxcnt = self.config.getint('structure', 'maxcnt')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.maxcnt = 50
            if self.maxcnt < 1:
                raise ValueError('maxcnt must be positive int')

    def _spgtuple(self):
        tmpspg = []
        for c in self.spgnum.split():
            if '-' in c:
                if not len(c.split('-')) == 2:
                    raise ValueError('Wrong input in spgnum.')
                istart = int(c.split('-')[0])
                iend = int(c.split('-')[1])+1
                if istart < 0 or 230 < istart:
                    raise ValueError('spgnum must be 1 -- 230')
                if iend < 0 or 231 < iend:
                    raise ValueError('spgnum must be 1 -- 230')
                for i in range(istart, iend):
                    if i not in tmpspg:
                        tmpspg.append(i)
            else:
                if int(c) < 0 or 230 < int(c):
                    raise ValueError('spgnum must be 1 -- 230')
                if not int(c) in tmpspg:
                    tmpspg += [int(c)]
        self.spgnum = tuple(tmpspg)

    def _read_option(self):
        # ---------- stop_chkpt
        try:
            self.stop_chkpt = self.config.getint('option', 'stop_chkpt')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.stop_chkpt = 0
        # ---------- load_struc_flag
        try:
            self.load_struc_flag = self.config.getboolean('option', 'load_struc_flag')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.load_struc_flag = False
        # ---------- stop_next_struc
        try:
            self.stop_next_struc = self.config.getboolean('option', 'stop_next_struc')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.stop_next_struc = False
        # ---------- recalc
        try:
            self.recalc = self.config.get('option', 'recalc')
            self.recalc = tuple([int(x) for x in self.recalc.split()])    # str --> int --> list --> tuple
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.recalc = None
        if self.recalc is not None:
            for cid in self.recalc:
                if not 0 <= cid < self.tot_struc:
                    raise ValueError('recalc must be non-negative int and less than tot_struc')
        # ---------- append_struc_ea
        try:
            self.append_struc_ea = self.config.getboolean('option', 'append_struc_ea')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.append_struc_ea = False
        # ---------- energy_step_flag
        try:
            self.energy_step_flag = self.config.getboolean('option', 'energy_step_flag')
            if self.calc_code in ['LAMMPS', 'OMX', 'ASE', 'ext']:
                raise ValueError('energy_step_flag: only VASP, QE, and soiap for now')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.energy_step_flag = False
        try:
            self.struc_step_flag = self.config.getboolean('option', 'struc_step_flag')
            if self.calc_code in ['LAMMPS', 'OMX', 'ASE', 'ext']:
                raise ValueError('struc_step_flag: only VASP, QE, and soiap for now')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.struc_step_flag = False
        try:
            self.force_step_flag = self.config.getboolean('option', 'force_step_flag')
            if self.calc_code in ['LAMMPS', 'OMX', 'ASE', 'ext']:
                raise ValueError('force_step_flag: only VASP, QE, and soiap for now')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.force_step_flag = False
        if self.algo == 'LAQA':
            self.force_step_flag = True
        try:
            self.stress_step_flag = self.config.getboolean('option', 'stress_step_flag')
            if self.calc_code in ['LAMMPS', 'OMX', 'ASE', 'ext']:
                raise ValueError('stress_step_flag: only VASP, QE, and soiap for now')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.stress_step_flag = False
        if self.algo == 'LAQA':
            self.stress_step_flag = True

    def _read_bo(self):
        # ---------- check physbo
        try:
            import physbo
        except ModuleNotFoundError:
            raise ModuleNotFoundError('PHYSBO is required for BO. --> pip3 install physbo')
        # ---------- nselect_bo
        self.nselect_bo = self.config.getint('BO', 'nselect_bo')
        if self.nselect_bo < 1:
            raise ValueError('nselect_bo < 1, check nselect_bo')
        elif self.tot_struc < self.nselect_bo:
            raise ValueError('tot_struc < nselect_bo, check nselect_bo')
        # ---------- score
        self.score = self.config.get('BO', 'score')
        if self.score in ['TS', 'EI', 'PI']:
            pass
        else:
            raise ValueError('score must be TS, EI, or PI')
        # ---------- num_rand_basis
        try:
            self.num_rand_basis = self.config.getint('BO', 'num_rand_basis')
        except configparser.NoOptionError:
            self.num_rand_basis = 0
        # ---------- cdev
        try:
            self.cdev = self.config.getfloat('BO', 'cdev')
        except configparser.NoOptionError:
            self.cdev = 0.001
        # ---------- dscrpt
        self.dscrpt = self.config.get('BO', 'dscrpt')
        # ---------- FP
        if self.dscrpt == 'FP':
            # ------ ValleOganov
            try:
                from dscribe.descriptors import ValleOganov
            except ModuleNotFoundError:
                raise ModuleNotFoundError('DScribe is required for FP. --> pip3 install dscribe')
            # ------ fp_rmax
            try:
                self.fp_rmax = self.config.getfloat('BO', 'fp_rmax')
            except configparser.NoOptionError:
                self.fp_rmax = 8.0
            if self.fp_rmax < 0:
                raise ValueError('fp_rmax < 0, check fp_rmin and fp_rmax')
            # ------ fp_npoints
            try:
                self.fp_npoints = self.config.getint('BO', 'fp_npoints')
            except configparser.NoOptionError:
                self.fp_npoints = 20
            if self.fp_npoints < 1:
                raise ValueError('fp_npoints < 1, check fp_npoints')
            # ------ fp_sigma
            try:
                self.fp_sigma = self.config.getfloat('BO', 'fp_sigma')
            except configparser.NoOptionError:
                self.fp_sigma = 0.7
            if self.fp_sigma < 0:
                raise ValueError('fp_sigma < 0, check fp_sigma')
        else:
            raise ValueError('dscrpt must be FP for now')
        # ---------- max_select_bo
        try:
            self.max_select_bo = self.config.getint('BO', 'max_select_bo')
        except configparser.NoOptionError:
            self.max_select_bo = 0
        if self.max_select_bo < 0:
            raise ValueError('max_select_bo < 0, check max_select_bo')
        # ---------- manual_select_bo
        try:
            self.manual_select_bo = self.config.get('BO', 'manual_select_bo')
            self.manual_select_bo = [int(x) for x in self.manual_select_bo.split()]
            self.manual_select_bo = tuple(self.manual_select_bo)    # list --> tuple
        except configparser.NoOptionError:
            self.manual_select_bo = None
        if self.manual_select_bo is not None:
            for cid in self.manual_select_bo:
                if not 0 <= cid < self.tot_struc:
                    raise ValueError('manual_select_bo must be non-negative int and less than tot_struc')
        # ---------- emin_bo, emax_bo
        try:
            self.emin_bo = self.config.getfloat('BO', 'emin_bo')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.emin_bo = None
        try:
            self.emax_bo = self.config.getfloat('BO', 'emax_bo')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.emax_bo = None
        if self.emax_bo is not None and self.emin_bo is not None:
            if self.emin_bo > self.emax_bo:
                raise ValueError('emax_bo < emin_bo, check emax_bo and emin_bo')

    def _read_laqa(self):
        # ---------- nselect_laqa
        self.nselect_laqa = self.config.getint('LAQA', 'nselect_laqa')
        if self.nselect_laqa < 1:
            raise ValueError('nselect_laqa < 1, check nselect_laqa')
        elif self.tot_struc < self.nselect_laqa:
            raise ValueError('tot_struc < nselect_laqa, check nselect_laqa')
        # ---------- wf
        try:
            self.wf = self.config.getfloat('LAQA', 'wf')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.wf = 0.1
        # ---------- ws
        try:
            self.ws = self.config.getfloat('LAQA', 'ws')
        except (configparser.NoOptionError, configparser.NoSectionError):
            self.ws = 10.0

    def _read_ea(self):
        if self.algo in ['EA', 'EA-vc'] or self.append_struc_ea:
            # ---------- n_pop
            self.n_pop = self.config.getint('EA', 'n_pop')
            if self.n_pop < 1:
                raise ValueError('n_pop must be positive int')
            # ---------- n_crsov
            self.n_crsov = self.config.getint('EA', 'n_crsov')
            if self.n_crsov < 0:
                raise ValueError('n_crsov must be non-negative int')
            # ---------- n_perm
            self.n_perm = self.config.getint('EA', 'n_perm')
            if self.n_perm < 0:
                raise ValueError('n_perm must be non-negative int')
            if self.n_perm != 0 and len(self.atype) == 1:
                raise ValueError('When the number of atom type is 1, n_perm must be 0')
            # ---------- n_strain
            self.n_strain = self.config.getint('EA', 'n_strain')
            if self.n_strain < 0:
                raise ValueError('n_strain must be non-negative int')
            # ---------- n_rand
            self.n_rand = self.config.getint('EA', 'n_rand')
            if self.n_rand < 0:
                raise ValueError('n_rand must be non-negative int')
            # ---------- check n_pop for EA
            if self.algo == 'EA' and self.struc_mode not in ['mol', 'mol_bs']:
                if self.n_crsov + self.n_perm + self.n_strain + self.n_rand != self.n_pop:
                    raise ValueError('n_crsov + n_perm + n_strain + n_rand must be n_pop')
            # ---------- n_elite
            self.n_elite = self.config.getint('EA', 'n_elite')
            if self.n_elite < 0:
                raise ValueError('n_elite must be non-negative int')
            # ---------- fit_reverse
            try:
                self.fit_reverse = self.config.getboolean('EA', 'fit_reverse')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.fit_reverse = False
            # ---------- n_fittest
            try:
                self.n_fittest = self.config.getint('EA', 'n_fittest')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.n_fittest = 0
            if self.n_fittest < 0:
                raise ValueError('n_fittest must be non-negative int')
            # ---------- slct_func
            self.slct_func = self.config.get('EA', 'slct_func')
            if self.slct_func not in ['TNM', 'RLT']:
                raise ValueError('slct_func must be TNM or RLT')
            # ------ TNM
            if self.slct_func == 'TNM':
                # -- t_size
                try:
                    self.t_size = self.config.getint('EA', 't_size')
                except (configparser.NoOptionError, configparser.NoSectionError):
                    self.t_size = 3
                if self.t_size < 2:
                    raise ValueError('t_size must be greater than or equal to 2')
            # ------ RLT
            elif self.slct_func == 'RLT':
                # -- a_rlt
                try:
                    self.a_rlt = self.config.getfloat('EA', 'a_rlt')
                except (configparser.NoOptionError, configparser.NoSectionError):
                    self.a_rlt = 10.0
                # -- b_rlt
                try:
                    self.b_rlt = self.config.getfloat('EA', 'b_rlt')
                except (configparser.NoOptionError, configparser.NoSectionError):
                    self.b_rlt = 1.0
                # -- check
                if not 0 < self.b_rlt < self.a_rlt:
                    raise ValueError('must be 0 < b_rlt < a_rlt')
            # ---------- crossover
            if self.n_crsov > 0:
                # ------ crs_lat
                try:
                    self.crs_lat = self.config.get('EA', 'crs_lat')
                except (configparser.NoOptionError, configparser.NoSectionError):
                    self.crs_lat = 'random'
                if self.crs_lat not in ['equal', 'random']:
                    raise ValueError('crs_lat must be equal or random')
                # ------ nat_diff_tole
                try:
                    self.nat_diff_tole = self.config.getint('EA', 'nat_diff_tole')
                except (configparser.NoOptionError, configparser.NoSectionError):
                    self.nat_diff_tole = 4
                if self.nat_diff_tole < 0:
                    raise ValueError('nat_diff_tole must be non-negative int')
            # ---------- permutation
            if self.n_perm > 0:
                # ------ ntimes
                try:
                    self.ntimes = self.config.getint('EA', 'ntimes')
                except (configparser.NoOptionError, configparser.NoSectionError):
                    self.ntimes = 1
                if self.ntimes < 1:
                    raise ValueError('ntimes must be positive int')
            # ---------- strain
            if self.n_strain > 0:
                # ------ sigma_st
                try:
                    self.sigma_st = self.config.getfloat('EA', 'sigma_st')
                except (configparser.NoOptionError, configparser.NoSectionError):
                    self.sigma_st = 0.5
                if self.sigma_st <= 0:
                    raise ValueError('sigma_st must be positive float')
            # ---------- other params
            # ------ maxcnt_ea
            try:
                self.maxcnt_ea = self.config.getint('EA', 'maxcnt_ea')
            except configparser.NoOptionError:
                self.maxcnt_ea = 50
            # ------ maxgen_ea
            try:
                self.maxgen_ea = self.config.getint('EA', 'maxgen_ea')
            except configparser.NoOptionError:
                self.maxgen_ea = 0
            if self.maxgen_ea < 0:
                raise ValueError('maxgen_ea must be non-negative int')
            # ------ emin_ea, emax_ea
            try:
                self.emin_ea = self.config.getfloat('EA', 'emin_ea')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.emin_ea = None
            try:
                self.emax_ea = self.config.getfloat('EA', 'emax_ea')
            except (configparser.NoOptionError, configparser.NoSectionError):
                self.emax_ea = None
            if self.emax_ea is not None and self.emin_ea is not None:
                if self.emin_ea > self.emax_ea:
                    raise ValueError('emax_ea < emin_ea, check emax_ea and emin_ea')
            # ---------- EA-vc
            if self.algo == 'EA-vc':
                # ------ n_add
                self.n_add = self.config.getint('EA', 'n_add')
                if self.n_add < 0:
                    raise ValueError('n_add must be non-negative int')
                # ------ n_elim
                self.n_elim = self.config.getint('EA', 'n_elim')
                if self.n_elim < 0:
                    raise ValueError('n_elim must be non-negative int')
                # ------ n_subs
                self.n_subs = self.config.getint('EA', 'n_subs')
                if self.n_subs < 0:
                    raise ValueError('n_subs must be non-negative int')
                # ------ target
                self.target = self.config.get('EA', 'target')
                #if self.target not in ['random','depop','overpop']:
                if self.target not in ['random']:
                    raise ValueError('target must be random for now')
                # ------ check n_pop for EA-vc
                if self.n_crsov + self.n_perm + self.n_strain + self.n_rand \
                    + self.n_add + self.n_elim + self.n_subs != self.n_pop:
                    raise ValueError('n_crsov + n_perm + n_strain + n_rand'
                                     ' + n_add + n_elim + n_subs must be n_pop')
                # ------ end_point
                self.end_point = self.config.get('EA', 'end_point')
                self.end_point = tuple([float(x) for x in self.end_point.split()])
                if not len(self.end_point) == len(self.atype):
                    raise ValueError('len(end_point) == len(atype), check end_point')
            # ---------- mol or mol_bs
            if self.struc_mode in ['mol', 'mol_bs']:
                # ------ n_rotation
                self.n_rotation = self.config.getint('EA', 'n_rotation')
                if self.n_rotation < 0:
                    raise ValueError('n_rotation must be non-negative int')
                # ------ check n_pop for mol or mol_bs
                if self.n_crsov + self.n_perm + self.n_strain + self.n_rand \
                    + self.n_rotation != self.n_pop:
                    raise ValueError('n_crsov + n_perm + n_strain + n_rand + n_rotation must be n_pop')
                # ------ mindist_mol_ea
                self.mindist_mol_ea = []
                for i in range(len(self.mol_file)):
                    tmp = self.config.get('EA', f'mindist_mol_ea_{i+1}')
                    tmp = tuple([float(x) for x in tmp.split()])    # str --> float
                    if not len(tmp) == len(self.mol_file):
                        raise ValueError(f'not len(mindist_mol_ea_{i+1}) == len(mol_file)')
                    self.mindist_mol_ea.append(tmp)
                self.mindist_mol_ea = tuple(self.mindist_mol_ea)    # list --> tuple
                # check symmetric matrix
                for i in range(len(self.mindist_mol_ea)):
                    for j in range(len(self.mindist_mol_ea)):
                        if i < j:
                            if not self.mindist_mol_ea[i][j] == self.mindist_mol_ea[j][i]:
                                raise ValueError(f'mindist_mol_ea is not symmetric. ({i}, {j}) -->'
                                                 f' {self.mindist_mol_ea[i][j]}, ({j}, {i}) --> {self.mindist_mol_ea[j][i]}')
                # ------ rot_max_angle
                try:
                    self.rot_max_angle = self.config.getint('EA', 'rot_max_angle')
                except configparser.NoOptionError:
                    self.rot_max_angle = 360
                if self.rot_max_angle <= 0:
                    raise ValueError('rot_max_angle must be positive int')
                # ------ protect_mol_struc
                try:
                    self.protect_mol_struc = self.config.getboolean('EA', 'protect_mol_struc')
                except configparser.NoOptionError:
                    self.protect_mol_struc = False

    def _read_vasp(self):
        # ---------- kpt_flag
        self.kpt_flag = True
        # ---------- kppvol
        self.kppvol = self.config.get('VASP', 'kppvol')
        self.kppvol = tuple([int(x) for x in self.kppvol.split()])    # str --> int --> list --> tuple
        if not len(self.kppvol) == self.nstage:
            raise ValueError('not len(kppvol) == nstage, check kppvol')
        # ---------- force_gamma
        try:
            self.force_gamma = self.config.getboolean('VASP', 'force_gamma')
        except configparser.NoOptionError:
            self.force_gamma = False

    def _read_qe(self):
        # ---------- kpt_flag
        self.kpt_flag = True
        # ---------- kppvol
        self.kppvol = self.config.get('QE', 'kppvol')
        self.kppvol = tuple([int(x) for x in self.kppvol.split()])    # str --> int --> list --> tuple
        if not len(self.kppvol) == self.nstage:
            raise ValueError('not len(kppvol) == nstage, check kppvol')
        # ---------- force_gamma
        try:
            self.force_gamma = self.config.getboolean('QE', 'force_gamma')
        except configparser.NoOptionError:
            self.force_gamma = False
        # ---------- qe_infile, qe_outfile
        self.qe_infile = self.config.get('QE', 'qe_infile')
        self.qe_outfile = self.config.get('QE', 'qe_outfile')
        # ---------- pv_term
        try:
            self.pv_term = self.config.getboolean('QE', 'pv_term')
        except configparser.NoOptionError:
            self.pv_term = False
        if self.energy_step_flag and self.pv_term:
            raise ValueError('cannot parse energy_step with pv_term yet. use pv_term = False')

    def _read_omx(self):
        # ---------- kpt_flag
        self.kpt_flag = False
        # ---------- kppvol
        self.kppvol = self.config.get('OMX', 'kppvol')
        self.kppvol = tuple([int(x) for x in self.kppvol.split()])    # str --> int --> list --> tuple
        if not len(self.kppvol) == self.nstage:
            raise ValueError('not len(kppvol) == nstage, check kppvol')
        # ---------- force_gamma
        try:
            self.force_gamma = self.config.getboolean('OMX', 'force_gamma')
        except configparser.NoOptionError:
            self.force_gamma = False
        # ---------- OMX_infile, OMX_outfile
        self.OMX_infile = self.config.get('OMX', 'OMX_infile')
        self.OMX_outfile = self.config.get('OMX', 'OMX_outfile')
        # ---------- upSpin, downSpin
        self.upSpin = {}
        self.downSpin = {}
        ValenceElec = self.config.get('OMX', 'ValenceElectrons')
        ValElecIn = ValenceElec.split()
        for i in range(0, len(ValElecIn), 3):
            self.upSpin[ValElecIn[i]]   = ValElecIn[i+1]
            self.downSpin[ValElecIn[i]] = ValElecIn[i+2]

    def _read_soiap(self):
        # ---------- kpt_flag
        self.kpt_flag = False
        # ---------- force_gamma
        self.force_gamma = False
        # ---------- soiap_infile, soiap_outfile, soiap_cif
        self.soiap_infile = self.config.get('soiap', 'soiap_infile')
        self.soiap_outfile = self.config.get('soiap', 'soiap_outfile')
        self.soiap_cif = self.config.get('soiap', 'soiap_cif')

    def _read_lammps(self):
        # ---------- kpt_flag
        self.kpt_flag = False
        # ---------- force_gamma
        self.force_gamma = False
        # ---------- lammps_infile, lammps_outfile
        self.lammps_infile = self.config.get('LAMMPS', 'lammps_infile')
        self.lammps_outfile = self.config.get('LAMMPS', 'lammps_outfile')
        # ---------- lammps_potential
        try:
            self.lammps_potential = self.config.get('LAMMPS', 'lammps_potential')
            self.lammps_potential = self.lammps_potential.split()
        except configparser.NoOptionError:
            self.lammps_potential = None
        # ---------- lammps_data
        self.lammps_data = self.config.get('LAMMPS', 'lammps_data')

    def _read_ase(self):
        # ---------- kpt_flag
        self.kpt_flag = False
        # ---------- force_gamma
        self.force_gamma = False
        # ---------- ase_python
        self.ase_python = self.config.get('ASE', 'ase_python')

    def _read_ext(self):
        # ---------- kpt_flag
        self.kpt_flag = False



from logging import getLogger

import numpy as np
from pymatgen.core import Structure

from ...util.struc_util import check_distance, rot_mat, find_site, cal_g, sort_by_atype_mol


logger = getLogger('cryspy')

# ea_gerenation.pyという古いコードから持ってきてまだ修正してない

#     def gen_rotation(self, rin, struc_mol_id, rot):
#         '''
#         generate structures by rotation
#         only for mol

#         # ---------- args
#         st: instance of rotation class
#         '''
#         # ---------- generate structures by rotation
#         struc_cnt = 0
#         while struc_cnt < rin.n_rotation:
#             # ------ select parents
#             pid, = self.sp.get_parents(rin=rin, n_parent=1)    # comma for list[0]
#             # ------ generate child
#             child, mol_id = rot.gen_child(rin, self.sp.struc_data[pid], struc_mol_id[pid])
#             # ------ success
#             if child is not None:
#                 self.offspring[self.cid] = child
#                 self.offspring_mol_id[self.cid] = mol_id
#                 self.parents[self.cid] = (pid, )    # tuple
#                 self.operation[self.cid] = 'rotation'
#                 try:
#                     spg_sym, spg_num = child.get_space_group_info(symprec=rin.symprec)
#                 except TypeError:
#                     spg_num = 0
#                     spg_sym = None
#                 logger.info(f'Structure ID {self.cid:>6} was generated'
#                       f' from {pid:>6} by rotation.'
#                       f' Space group: {spg_num:>3} {spg_sym}')
#                 self.cid += 1
#                 struc_cnt += 1




# クラスから関数に変更予定


# class Rotation:
#     '''
#     rotation
#     '''

#     def __init__(self, mindist):
#         self.mindist = mindist

#     def gen_child(self, rin, struc, mol_id):
#         # ---------- init
#         cnt = 0
#         self.child = struc
#         real_coords, real_group_id, real_species, _ = find_site(self.child, mol_id[0], mol_id[1], mol_id[2])
#         mol_gs = self.child.lattice.get_cartesian_coords(cal_g(self.child, mol_id[0], mol_id[1], mol_id[2]))
#         rotated_group_id = []
#         # ---------- generate rotated structure
#         while True:
#             cnt += 1
#             rot_coords = []
#             rot_species = []
#             # ------ rotation molecules
#             for i in range(sum(rin.nmol)):
#                 # -- calculate rotation matrix
#                 mol_angle = np.deg2rad(rin.rot_max_angle) * np.random.rand(3)
#                 R = rot_mat(mol_angle)
#                 # -- rotate
#                 for j in range(len(self.child)):
#                     if real_group_id[j] == i:
#                         rot_coords.append(np.matmul(self.child.lattice.get_cartesian_coords(real_coords[j]) -
#                                                     mol_gs[i], R) + mol_gs[i])
#                         rot_species.append(real_species[j])
#                         rotated_group_id.append(i)
#             # ------ child
#             self.child = Structure(lattice=self.child.lattice, species=rot_species, coords=rot_coords,
#                                    coords_are_cartesian=True)

#             # ------ check distance
#             success, mindist_ij, dist = check_distance(self.child,
#                                                        rin.atype,
#                                                        self.mindist)
#             if success:
#                 self.child, self.mol_id, self.group_id = sort_by_atype_mol(self.child, rin.atype,
#                                                                            mol_id[0], rotated_group_id)
#                 return self.child, [self.mol_id, self.group_id, mol_id[2]]
#             else:
#                 type0 = rin.atype[mindist_ij[0]]
#                 type1 = rin.atype[mindist_ij[1]]
#                 logger.warning(f'mindist in rotation: {type0} - {type1}, {dist}. retry.')
#                 cnt += 1
#                 if cnt >= rin.maxcnt_ea:
#                     logger.warning('Rotation: could not satisfy min_dist' +
#                           f' in {rin.maxcnt_ea} times')
#                     logger.warning('Change parent')
#                     self.child = None
#                     return None, None    # change parent



'''
Output results in ./data/xxx
'''


def out_rslt(rslt_data, order_ef=False):
    # ---------- asc in Struc_ID or (Gen or Select)
    with open('./data/cryspy_rslt', 'w') as f:
        f.write(rslt_data.to_string())

    # ---------- asc in energy
    with open('./data/cryspy_rslt_energy_asc', 'w') as f:
        if not order_ef:
            order = 'E_eV_atom'
        else:
            order = 'Ef_eV_atom'
        f.write(rslt_data.sort_values(
            by=[order], ascending=True).to_string())


def out_kpts(kpt_data):
    # ------ asc in ID
    with open('./data/kpts_rslt', 'w') as f:
        f.write(f'{"Struc_ID":>10}  {"k-points":>10}\n')
        for key, value in sorted(kpt_data.items()):
            f.write(f'{key:10d}  {value}\n')


# ---------- BO
def out_bo_status(bo_mean, bo_var, bo_score, n_selection):
    with open('./data/bo_status', 'w') as f:
        # ------ label
        f.write(f'{"Struc_ID":>10}{"Score":>14}{"Mean":>14}{"Variance":>14}\n')
        # ------ sorted by score
        for cid, value in sorted(bo_score[n_selection].items(),
                                 key=lambda x: -x[1]):
            f.write(f'{cid:>10d}')
            f.write(f'{value:>14.8f}')
            f.write(f'{bo_mean[n_selection][cid]:>14.8f}')
            f.write(f'{bo_var[n_selection][cid]:>14.8f}')
            f.write('\n')


def out_bo_common(bo_str, bo_dict, tot_struc):
    len_select = len(bo_dict)
    with open('./data/'+bo_str, 'w') as f:
        # ------ label
        f.write('  Struc_ID')
        for i in range(len_select):
            f.write(f'    Select {i+1:<3d}')
        f.write('\n')
        # ------ values
        for cid in range(tot_struc):
            f.write(f'{cid:>10d}')
            for i in range(2, len_select + 2):    # start from 2
                if cid in bo_dict[i]:
                    f.write(f'{bo_dict[i][cid]:14.8f}')
                else:
                    f.write('              ')
            f.write('\n')


def out_bo_id_hist(id_select_hist):
    with open('./data/bo_select_id', 'w') as f:
        f.write(f'{"Selection":>10}  {"ID":>5}\n')
        for i, j in enumerate(id_select_hist):
            f.write(f'{i+1:10d}')
            for x in j:
                f.write(f'  {x:5d}')
            f.write('\n')


# ---------- LAQA
def out_laqa_status(laqa_step, laqa_score, laqa_energy, laqa_bias):
    # ------ desc in score
    with open('./data/laqa_status', 'w') as f:
        f.write(f'{"Struc_ID":>10}  {"Score":>14}  {"E_eV_atom":>14}'
                f'  {"Bias":>14}  {"Selection":>12}  {"Step":>12}\n')
        for key, value in sorted(laqa_score.items(), key=lambda x: -x[1][-1]):
            if laqa_energy[key]:    # whether list is vacant or not?
                f.write(f'{key:10d}  {value[-1]: 14.5f}  {laqa_energy[key][-1]: 14.5f}'
                        f'  {laqa_bias[key][-1]: 14.5f}  {len(laqa_step[key]):12d}'
                        f'  {sum(laqa_step[key]):12d}\n')
            else:
                f.write(f'{key:10d}  {value[-1]: 14.5f}  {"":>14}'
                        f'  {"":>14}  {len(laqa_step[key]):12d}  {sum(laqa_step[key]):12d}\n')


def out_laqa_step(laqa_step):
    # ------ asc in ID
    with open('./data/laqa_step', 'w') as f:
        f.write(f'{"Struc_ID":>10}  {"Step":>4}\n')
        for key, value in sorted(laqa_step.items()):
            f.write(f'{key:10d}')
            for x in value:
                f.write(f'  {x:4d}')
            f.write('\n')


def out_laqa_score(laqa_score):
    # ------ asc in ID
    with open('./data/laqa_score', 'w') as f:
        f.write(f'{"Struc_ID":>10}  {"Score":>14}\n')
        for key, value in sorted(laqa_score.items()):
            f.write(f'{key:10d}')
            for x in value:
                f.write(f'  {x: 14.5f}')
            f.write('\n')


def out_laqa_energy(laqa_energy):
    # ------ asc in ID
    with open('./data/laqa_energy', 'w') as f:
        f.write(f'{"Struc_ID":>10}  {"E(eV/atom)":>12}\n')
        for key, value in sorted(laqa_energy.items()):
            f.write(f'{key:10d}')
            for x in value:
                f.write(f'  {x: 12.5f}')
            f.write('\n')


def out_laqa_bias(laqa_bias):
    # ------ asc in ID
    with open('./data/laqa_bias', 'w') as f:
        f.write(f'{"Struc_ID":>10}  {"Bias":>14}\n')
        for key, value in sorted(laqa_bias.items()):
            f.write(f'{key:10d}')
            for x in value:
                f.write(f'  {x: 14.5f}')
            f.write('\n')


def out_laqa_id_hist(id_select_hist):
    with open('./data/laqa_select_id', 'w') as f:
        f.write(f'{"Selection":>10}  {"ID":>5}\n')
        for i, j in enumerate(id_select_hist):
            f.write(f'{i+1:10d}')
            for x in j:
                f.write(f'  {x:5d}')
            f.write('\n')


# ---------- EA
def out_ea_info(ea_info):
    with open('./data/ea_info', 'w') as f:
        f.write(ea_info.to_string(index=False))


def out_ea_origin(ea_origin):
    with open('./data/ea_origin', 'w') as f:
        f.write(ea_origin.to_string(index=False))


# ---------- EA-vc
def out_nat_data(nat_data, atype):
    with open('./data/nat_data', 'w') as f:
        f.write(f'    ID    {atype}\n')
        for cid, nat in nat_data.items():
            f.write(f'{cid:6}    {nat}\n')


def out_hdist(gen, hdist, ratio_data):
    with open(f'./data/convex_hull/hull_dist_all_gen_{gen}', 'w') as f:
        f.write(f'    ID    hull distance (eV/atom)    ratio\n')
        for cid, dist in sorted(hdist.items(), key=lambda x: x[1]):
            f.write(f'{cid:6}    {dist:>23.6f}    {ratio_data[cid]}\n')


from logging import getLogger

import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import Structure

from ...util.struc_util import sort_by_atype, check_distance, cal_g, sort_by_atype_mol, find_site


logger = getLogger('cryspy')


def gen_permutation(
        atype,
        mindist,
        struc_data,
        sp,
        n_perm,
        id_start=None,
        symprec=0.01,
        ntimes=1,
        maxcnt_ea=50,
        struc_mol_id=None,
        molecular=False,
    ):
    '''

        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    mindist (tuple): minimum interatomic distance, e.g. ((1.5, 1.5), (1.5, 1.5))
    struc_data (dict): {id: structure data}
    sp (instance): instance of SelectParents class
    n_perm (int): number of structures to generate by permutation
    id_start (int): start ID for new structures
    symprec (float): tolerance for symmetry
    ntimes (int): number of swaps
    maxcnt_ea (int): maximum number of trial in permutation

    # ---------- return
    children (dict): {id: structure data}
    parents (dict): {id: (id of parent_A, )}
    operation (dict): {id: 'permutation'}
    '''

    # ---------- initialize
    struc_cnt = 0
    children = {}
    #children_mol_id = {}
    parents = {}
    operation = {}

    # ---------- id_offset
    if id_start is None:
        cid = max(struc_data.keys()) + 1
    else:
        if id_start < (max(struc_data.keys()) + 1):
            logger.error('id_start is already included in structure ID of the data')
        else:
            cid = id_start

    # ---------- generate structures by permutation
    while struc_cnt < n_perm:
        # ------ select parents
        pid_A, = sp.get_parents(n_parent=1)    # comma for list[0]
        parent_A = struc_data[pid_A]
        # ------ generate child
        if molecular:
            logger.error('Permutation for molecular is not implemented yet')
            # child, mol_id = pm.gen_child_mol(rin,
            #                                     self.sp.struc_data[pid_A],
            #                                     struc_mol_id[pid_A])
        else:
            child = gen_child(atype, mindist, parent_A, ntimes, maxcnt_ea)
        # ------ success
        if child is not None:
            children[cid] = child
            # if molecular:
            #     children_mol_id[cid] = mol_id
            parents[cid] = (pid_A, )    # tuple
            operation[cid] = 'permutation'
            try:
                spg_sym, spg_num = child.get_space_group_info(symprec=symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
            logger.info(f'Structure ID {cid:>6} was generated'
                    f' from {pid_A:>6} by permutation.'
                    f' Space group: {spg_num:>3} {spg_sym}')
            cid += 1
            struc_cnt += 1

    # ---------- return
    return children, parents, operation


def gen_child(atype, mindist, parent_A, ntimes=1, maxcnt_ea=50):
    '''

        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    mindist (tuple): minimum interatomic distance, e.g. ((1.5, 1.5), (1.5, 1.5))
    parent_A (Structure): pymatgen Structure object
    ntimes (int): number of swaps
    maxcnt_ea (int): maximum number of trial in crossover

    # ---------- return
    (if success) child (Structure): pymatgen Structure object
    (if fail) None
    '''

    # ---------- initialize
    smatcher = StructureMatcher()    # instantiate StructureMatcher
    cnt = 0

    # ---------- ntimes permutation
    while True:
        child = parent_A.copy()    # keep original structure
        n = ntimes
        while n > 0:
            # ------ prepare index for each atom type
            indx_each_type = []
            for a in atype:
                indx_each_type.append(
                    [i for i, site in enumerate(child)
                        if site.species_string == a])
            # ------ choose two atom type
            type_choice = np.random.choice(len(atype), 2,
                                            replace=False)
            # ------ choose index
            indx_choice = []
            for tc in type_choice:
                indx_choice.append(np.random.choice(indx_each_type[tc]))
            # ------ replace each other
            child.replace(indx_choice[0],
                                species=atype[type_choice[1]])
            child.replace(indx_choice[1],
                                species=atype[type_choice[0]])
            # ------ compare to original one
            if smatcher.fit(child, parent_A):
                n = ntimes    # back to the start
                continue
            else:
                n -= 1

        # ------ check distance
        success, mindist_ij, dist = check_distance(child, atype, mindist)
        if success:
            child = sort_by_atype(child, atype)
            return child
        else:
            type0 = atype[mindist_ij[0]]
            type1 = atype[mindist_ij[1]]
            logger.warning(f'mindist in permutation: {type0} - {type1}, {dist}. retry.')
            cnt += 1
            if cnt >= maxcnt_ea:
                logger.warning('Permutatin: could not satisfy min_dist' +
                        f' in {maxcnt_ea} times')
                logger.warning('Change parent')
                return None    # change parent



from logging import getLogger

import numpy as np
from pymatgen.core import Structure
from pymatgen.core.periodic_table import DummySpecie

from ...util.struc_util import sort_by_atype, sort_by_atype_mol, check_distance, cal_g, find_site


logger = getLogger('cryspy')


def gen_strain(
        atype,
        mindist,
        struc_data,
        sp,
        n_strain,
        id_start=None,
        symprec=0.01,
        sigma_st=0.5,
        maxcnt_ea=50,
        struc_mol_id=None,
        molecular=False,
        protect_mol_struc=True,
    ):
    '''

        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    mindist (tuple): minimum interatomic distance, e.g. ((1.5, 1.5), (1.5, 1.5))
    struc_data (dict): {id: structure data}
    sp (instance): instance of SelectParents class
    n_strain (int): number of structures to generate by strain
    id_start (int): start ID for new structures
    symprec (float): tolerance for symmetry
    sigma_st (float): standard deviation for strain matrix
    maxcnt_ea (int): maximum number of trial in permutation

    # ---------- return
    children (dict): {id: structure data}
    parents (dict): {id: (id of parent_A, )}
    operation (dict): {id: 'strain'}
    '''
    # ---------- initialize
    struc_cnt = 0
    children = {}
    #children_mol_id = {}
    parents = {}
    operation = {}

    # ---------- id_offset
    if id_start is None:
        cid = max(struc_data.keys()) + 1
    else:
        if id_start < (max(struc_data.keys()) + 1):
            logger.error('id_start is already included in structure ID of the data')
        else:
            cid = id_start

    # ---------- generate structures by strain
    while struc_cnt < n_strain:
        # ------ select parents
        pid_A, = sp.get_parents(n_parent=1)    # comma for list[0]
        parent_A = struc_data[pid_A]
        # ------ generate child
        if molecular:
            logger.error('protect_mol_struc is not implemented yet.')
            # if protect_mol_struc:
                # child, mol_id = gen_child_mol(rin,
                #                                     struc_data[pid],
                #                                     struc_mol_id[pid])
            #else:
            #    child = gen_child(atype, mindist, parent_struc, sigma_st, maxcnt_ea)
        else:
            child = gen_child(atype, mindist, parent_A, sigma_st, maxcnt_ea)
        # ------ success
        if child is not None:
            children[cid] = child
            # if molecular:
            #     children_mol_id[cid] = mol_id
            parents[cid] = (pid_A, )    # tuple
            operation[cid] = 'strain'
            try:
                spg_sym, spg_num = child.get_space_group_info(symprec=symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
            logger.info(f'Structure ID {cid:>6} was generated'
                    f' from {pid_A:>6} by strain.'
                    f' Space group: {spg_num:>3} {spg_sym}')
            cid += 1
            struc_cnt += 1

    # ---------- return
    return children, parents, operation


def gen_child(atype, mindist, parent_A, sigma_st=0.5, maxcnt_ea=50):
    '''

        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    mindist (tuple): minimum interatomic distance, e.g. ((1.5, 1.5), (1.5, 1.5))
    parent_A (Structure): pymatgen Structure object
    sigma_st (float): standard deviation for strain matrix
    maxcnt_ea (int): maximum number of trial in crossover

    # ---------- return
    (if success) child (Structure): pymatgen Structure object
    (if fail) None
    '''
    # ---------- initialize
    child = parent_A.copy()    # keep original structure
    lat_mat = child.lattice.matrix.T    # lattice vector as matrix
    cnt = 0

    # ---------- generate strained structure
    while True:
        # ------ strain matrix
        strain_matrix = np.empty([3, 3])
        for i in range(3):
            for j in range(3):
                if i <= j:
                    if i == j:
                        strain_matrix[i][j] = 1.0 + np.random.normal(loc=0.0, scale=sigma_st)
                    else:
                        strain_matrix[i][j] = np.random.normal(loc=0.0, scale=sigma_st)/2.0
                        strain_matrix[j][i] = strain_matrix[i][j]
        # ------ strained lattice
        strained_lattice = np.dot(strain_matrix, lat_mat).T
        # ------ child
        child = Structure(strained_lattice, child.species, child.frac_coords)
        # ------ scale lattice
        child.scale_lattice(parent_A.volume)
        # ------ check distance
        success, mindist_ij, dist = check_distance(child, atype, mindist)
        if success:
            child = sort_by_atype(child, atype)
            return child
        else:
            type0 = atype[mindist_ij[0]]
            type1 = atype[mindist_ij[1]]
            logger.warning(f'mindist in strain: {type0} - {type1}, {dist}. retry.')
            cnt += 1
            if cnt >= maxcnt_ea:
                return None    # change parent

